# ===========================================================
# ========================= imports =========================
import time
import pandas as _pd
import numpy as _np
import datetime as _dt
from typing import Union
from datetime import datetime, timedelta
from gnssvod.io import readFile
from gnssvod.geodesy.coordinate import cart2ell, geocentric_latitude
from gnssvod.funcs.filename import (sp3FileName, clockFileName, ionFileName)
from gnssvod.funcs.interpolation import coord_interp
from gnssvod.position.position import _observation_picker_by_band
# ===========================================================

__all__ = ["sp3_interp", "ionosphere_interp"]

def sp3_interp_fast(start_time, end_time, interval=30, poly_degree=16, sp3_product="gfz", clock_product="gfz", aux_path = Union[str, None]):
    # add a buffer around start and end time
    start_time_withbuff = start_time-_dt.timedelta(hours=2.1)
    end_time_withbuff = end_time+_dt.timedelta(hours=2.1)
    # determine all required orbit and clock files
    start_epoch = _dt.date(start_time_withbuff.year,start_time_withbuff.month,start_time_withbuff.day)
    end_epoch = _dt.date(end_time_withbuff.year,end_time_withbuff.month,end_time_withbuff.day)
    epochs = _pd.date_range(start_epoch,end_epoch,freq='d')
    sp3FileNames = [sp3FileName(x.date(),sp3_product,aux_path) for x in epochs]
    clockFileNames = [clockFileName(x.date(),interval,clock_product,aux_path) for x in epochs]
    #--------------------------------------------------------------------------
    # reading all those files, downloading them if necessary
    sp3 = [readFile.read_sp3File(x) for x in sp3FileNames]
    clock = [readFile.read_clockFile(x) for x in clockFileNames]
    #--------------------------------------------------------------------------
    # concatenate data from current day and some buffer before and after
    start = time.time()
    sp3 = _pd.concat(sp3,axis=0)
    clock = _pd.concat(clock,axis=0)
    sp3 = sp3.dropna(subset=["deltaT"]).reorder_levels(["Epoch","SV"])
    sp3 = sp3.loc[(slice(start_time_withbuff,end_time_withbuff))]
    #--------------------------------------------------------------------------
    svList = sp3.index.get_level_values("SV").unique()
    svList = svList.sort_values()
    #--------------------------------------------------------------------------
    # interpolate orbit parameters to the required time interval
    sp3_resampled = []
    for sv in svList:
        sp3_temp = sp3.xs(sv,level='SV')[['X','Y','Z']] * 1000 # km to m
        # only process if a minimum of 4 orbit data points are present
        if len(sp3_temp)>3:
            sp3_temp_resampled = sp3_temp.resample(f"{interval}s")
            sp3_temp_resampled = sp3_temp_resampled.interpolate(method='cubic')
            # recalculate V taking into account sampling rate
            sp3_temp_resampled['Vx']=-sp3_temp_resampled['X'].diff(periods=-1)/interval
            sp3_temp_resampled['Vy']=-sp3_temp_resampled['Y'].diff(periods=-1)/interval
            sp3_temp_resampled['Vz']=-sp3_temp_resampled['Z'].diff(periods=-1)/interval
            sp3_resampled.append(sp3_temp_resampled)

    sp3_resampled = _pd.concat(sp3_resampled, keys=svList, names=['SV']).reorder_levels(['Epoch', 'SV']).sort_index(level='Epoch')
    #--------------------------------------------------------------------------
    # interpolate clock parameters to the required time interval
    clock.index.name = 'SV'
    clock.set_index('Epoch', append=True, inplace=True)
    clock = clock.reorder_levels(['Epoch', 'SV'])
    svList_clk = clock.index.get_level_values("SV").unique()
    svList_clk = svList_clk.sort_values()
    
    clock_resampled = []
    for sv in svList_clk:
        clock_temp = clock.xs(sv,level='SV')
        # only process if a minimum of 4 clock data points are present
        if len(clock_temp)>3:
            clock_temp_resampled = clock_temp.resample(f"{interval}s")
            clock_temp_resampled = clock_temp_resampled.interpolate(method='cubic')
            clock_resampled.append(clock_temp_resampled)

    clock_resampled = _pd.concat(clock_resampled, keys=svList_clk, names=['SV']).reorder_levels(['Epoch', 'SV']).sort_index(level='Epoch')
    #--------------------------------------------------------------------------
    # finally, merge orbit and clock information
    sp3matched = sp3_resampled.join(clock_resampled,how='inner')
    '''
    # all of this commented script is the former sp3_interp function
    # it is kept here in case one wants to compare the output of the new implementation above
    # against the old implementation below
    #--------------------------------------------------------------------------
    epoch_values = sp3.index.get_level_values("Epoch").unique()
    deltaT = epoch_values[1]-epoch_values[0]
    fitTime = _np.linspace(0, deltaT.seconds*16 , 17)
    header = ['X', 'Y', 'Z', 'Vx','Vy','Vz']
    # --------------------------------------------------------------
    epoch_start = _pd.Timestamp(datetime(epoch_yesterday.year, epoch_yesterday.month, epoch_yesterday.day,23,0,0))
    epoch_step = timedelta(hours=3)
    epoch_stop = epoch_start + timedelta(hours=4)
    dti = _pd.date_range(start = _pd.Timestamp(datetime(epoch_yesterday.year, epoch_yesterday.month, epoch_yesterday.day,23,30,0)),
                        end   = _pd.Timestamp(datetime( epoch_tomorrow.year,  epoch_tomorrow.month,  epoch_tomorrow.day,2,29,59)),
                        freq = str(interval) + 'S') 
    index = _pd.MultiIndex.from_product([svList, dti.tolist()], names=['SV', 'Epoch'])
    interp_coord = _pd.DataFrame(index= index, columns = header)
    interp_coord = interp_coord.reorder_levels(['Epoch', 'SV'])
    interp_coord = interp_coord.sort_index()
    while True:
        sp3_temp = sp3.loc[(slice(epoch_start,epoch_stop))].copy()
        sp3_temp = sp3_temp.reorder_levels(["SV","Epoch"])
        epoch_interp_List = _np.zeros(shape=(int(10800/interval),6,len(svList)))
        for svIndex,sv in enumerate(svList):
            if sv in sp3_temp.index.get_level_values('SV'):
                epoch_number = len(sp3_temp.loc[sv])
            else:
                epoch_number = 0
            if epoch_number <= poly_degree:
                print("Warning: Not enough epochs to predict for satellite",sv,"| Epoch Count:",epoch_number, " - Polynomial Degree:",poly_degree)
                epoch_interp_List[:,:,svIndex] = _np.full(shape=(int(10800/interval),6),fill_value=None)
                continue
            if epoch_number != 17:
                fitTime = [(sp3_temp.loc[sv].index[t]-sp3_temp.loc[sv].index[0]).seconds for t in range(epoch_number)]
            # Fit sp3 coordinates to 16 deg polynomial
            fitX = _np.polyfit(fitTime, sp3_temp.loc[sv].X.copy(), deg=poly_degree)
            fitY = _np.polyfit(fitTime, sp3_temp.loc[sv].Y.copy(), deg=poly_degree)
            fitZ = _np.polyfit(fitTime, sp3_temp.loc[sv].Z.copy(), deg=poly_degree)
            # Interpolate coordinates
            x_interp = coord_interp(fitX, interval) * 1000 # km to m
            x_velocity = _np.array([(x_interp[i+1]-x_interp[i])/interval if (i+1)<len(x_interp) else 0 for i in range(len(x_interp))])
            y_interp = coord_interp(fitY, interval) * 1000 # km to m
            y_velocity = _np.array([(y_interp[i+1]-y_interp[i])/interval if (i+1)<len(y_interp) else 0 for i in range(len(y_interp))])
            z_interp = coord_interp(fitZ, interval) * 1000 # km to m
            z_velocity = _np.array([(z_interp[i+1]-z_interp[i])/interval if (i+1)<len(z_interp) else 0 for i in range(len(z_interp))])
            sv_interp = _np.vstack((x_interp[:-1], y_interp[:-1], z_interp[:-1], x_velocity[:-1], y_velocity[:-1], z_velocity[:-1])).transpose()
            epoch_interp_List[:,:,svIndex] = sv_interp
            fitTime = _np.linspace(0, deltaT.seconds*16 , 17) # restore original fitTime in case it has changed!
        interp_coord.loc[(slice(epoch_start+timedelta(minutes=30),epoch_stop-timedelta(minutes=30,seconds=1))), ('X', 'Y', 'Z', 'Vx', 'Vy', 'Vz')] = epoch_interp_List.transpose(1,0,2).reshape(6,-1).transpose()
        epoch_start += epoch_step
        epoch_stop += epoch_step
        if epoch_start == _pd.Timestamp(datetime(epoch_tomorrow.year, epoch_tomorrow.month, epoch_tomorrow.day,2,0,0)):
            break
    # -------------------------------------------------------------------------
    interp_coord = interp_coord.reorder_levels(['Epoch', 'SV']).astype(float)
    clock.index.name = 'SV'
    clock.set_index('Epoch', append=True, inplace=True)
    clock = clock.reorder_levels(['Epoch', 'SV'])
    epochMatch = interp_coord.index.intersection(clock.index) 
    ephemerisMatched = interp_coord.loc[epochMatch] 
    clockMatched   = clock.loc[epochMatch]
    sp3matched_old = _pd.concat([ephemerisMatched, clockMatched], axis=1)
    pdb.set_trace()
    # -------------------------------------------------------------------------
    '''
    finish = time.time()
    print("SP3 interpolation is done in", '{0:.2f}'.format(finish-start), 'seconds')
    return sp3matched

def sp3_interp(epoch, interval=30, poly_degree=16, sp3_product="gfz", clock_product="gfz"):
    epoch_yesterday = epoch - timedelta(days=1)
    epoch_tomorrow =  epoch + timedelta(days=1)
    yesterday = sp3FileName(epoch_yesterday, sp3_product)
    today = sp3FileName(epoch, sp3_product)
    tomorrow = sp3FileName(epoch_tomorrow, sp3_product)
    clockFile = clockFileName(epoch, interval, clock_product)
    #--------------------------------------------------------------------------
    yes   = readFile.read_sp3File(yesterday) 
    tod = readFile.read_sp3File(today) 
    tom   = readFile.read_sp3File(tomorrow) 
    clock = readFile.read_clockFile(clockFile)
    #--------------------------------------------------------------------------
    if poly_degree > 16:
        raise Warning("Polynomial degree above 16 is not applicable!")
    elif poly_degree < 11:
        print("Warning: Polynomial degree below 11 is not recommended!")
    #--------------------------------------------------------------------------
    start = time.time()
    yes = yes.dropna(subset=["deltaT"])
    tod = tod.dropna(subset=["deltaT"])
    tom = tom.dropna(subset=["deltaT"])
    yes = yes.reorder_levels(["Epoch","SV"])
    tod = tod.reorder_levels(["Epoch","SV"])
    tom = tom.reorder_levels(["Epoch","SV"])
    yes = yes.loc[(slice(_pd.Timestamp(datetime(epoch_yesterday.year, epoch_yesterday.month, epoch_yesterday.day,23,0,0)),_pd.Timestamp(datetime(epoch_yesterday.year, epoch_yesterday.month, epoch_yesterday.day,23,59,59))))]
    tom = tom.loc[(slice(_pd.Timestamp(datetime(epoch_tomorrow.year, epoch_tomorrow.month, epoch_tomorrow.day,0,0,0)),_pd.Timestamp(datetime(epoch_tomorrow.year, epoch_tomorrow.month, epoch_tomorrow.day,3,0,0))))]
    sp3 = _pd.concat([yes,tod,tom], axis=0)
    svList = sp3.index.get_level_values("SV").unique()
    svList = svList.sort_values()
    #--------------------------------------------------------------------------
    epoch_values = sp3.index.get_level_values("Epoch").unique()
    deltaT = epoch_values[1]-epoch_values[0]
    fitTime = _np.linspace(0, deltaT.seconds*16 , 17)
    header = ['X', 'Y', 'Z', 'Vx','Vy','Vz']
    # --------------------------------------------------------------
    epoch_start = _pd.Timestamp(datetime(epoch_yesterday.year, epoch_yesterday.month, epoch_yesterday.day,23,0,0))
    epoch_step = timedelta(hours=3)
    epoch_stop = epoch_start + timedelta(hours=4)
    dti = _pd.date_range(start = _pd.Timestamp(datetime(epoch_yesterday.year, epoch_yesterday.month, epoch_yesterday.day,23,30,0)),
                        end   = _pd.Timestamp(datetime( epoch_tomorrow.year,  epoch_tomorrow.month,  epoch_tomorrow.day,2,29,59)),
                        freq = str(interval) + 'S') 
    index = _pd.MultiIndex.from_product([svList, dti.tolist()], names=['SV', 'Epoch'])
    interp_coord = _pd.DataFrame(index= index, columns = header)
    interp_coord = interp_coord.reorder_levels(['Epoch', 'SV'])
    interp_coord = interp_coord.sort_index()
    while True:
        sp3_temp = sp3.loc[(slice(epoch_start,epoch_stop))].copy()
        sp3_temp = sp3_temp.reorder_levels(["SV","Epoch"])
        epoch_interp_List = _np.zeros(shape=(int(10800/interval),6,len(svList)))
        for svIndex,sv in enumerate(svList):
            if sv in sp3_temp.index.get_level_values('SV'):
                epoch_number = len(sp3_temp.loc[sv])
            else:
                epoch_number = 0
            if epoch_number <= poly_degree:
                print("Warning: Not enough epochs to predict for satellite",sv,"| Epoch Count:",epoch_number, " - Polynomial Degree:",poly_degree)
                epoch_interp_List[:,:,svIndex] = _np.full(shape=(int(10800/interval),6),fill_value=None)
                continue
            if epoch_number != 17:
                fitTime = [(sp3_temp.loc[sv].index[t]-sp3_temp.loc[sv].index[0]).seconds for t in range(epoch_number)]
            # Fit sp3 coordinates to 16 deg polynomial
            fitX = _np.polyfit(fitTime, sp3_temp.loc[sv].X.copy(), deg=poly_degree)
            fitY = _np.polyfit(fitTime, sp3_temp.loc[sv].Y.copy(), deg=poly_degree)
            fitZ = _np.polyfit(fitTime, sp3_temp.loc[sv].Z.copy(), deg=poly_degree)
            # Interpolate coordinates
            x_interp = coord_interp(fitX, interval) * 1000 # km to m
            x_velocity = _np.array([(x_interp[i+1]-x_interp[i])/interval if (i+1)<len(x_interp) else 0 for i in range(len(x_interp))])
            y_interp = coord_interp(fitY, interval) * 1000 # km to m
            y_velocity = _np.array([(y_interp[i+1]-y_interp[i])/interval if (i+1)<len(y_interp) else 0 for i in range(len(y_interp))])
            z_interp = coord_interp(fitZ, interval) * 1000 # km to m
            z_velocity = _np.array([(z_interp[i+1]-z_interp[i])/interval if (i+1)<len(z_interp) else 0 for i in range(len(z_interp))])
            sv_interp = _np.vstack((x_interp[:-1], y_interp[:-1], z_interp[:-1], x_velocity[:-1], y_velocity[:-1], z_velocity[:-1])).transpose()
            epoch_interp_List[:,:,svIndex] = sv_interp
            fitTime = _np.linspace(0, deltaT.seconds*16 , 17) # restore original fitTime in case it has changed!
        interp_coord.loc[(slice(epoch_start+timedelta(minutes=30),epoch_stop-timedelta(minutes=30,seconds=1))), ('X', 'Y', 'Z', 'Vx', 'Vy', 'Vz')] = epoch_interp_List.transpose(1,0,2).reshape(6,-1).transpose()
        epoch_start += epoch_step
        epoch_stop += epoch_step
        if epoch_start == _pd.Timestamp(datetime(epoch_tomorrow.year, epoch_tomorrow.month, epoch_tomorrow.day,2,0,0)):
            break
    # -------------------------------------------------------------------------
    interp_coord = interp_coord.reorder_levels(['Epoch', 'SV']).astype(float)
    clock.index.name = 'SV'
    clock.set_index('Epoch', append=True, inplace=True)
    clock = clock.reorder_levels(['Epoch', 'SV'])
    epochMatch = interp_coord.index.intersection(clock.index) 
    ephemerisMatched = interp_coord.loc[epochMatch] 
    clockMatched   = clock.loc[epochMatch]
    sp3matched = _pd.concat([ephemerisMatched, clockMatched], axis=1)
    # -------------------------------------------------------------------------
    finish = time.time()
    print("SP3 interpolation is done in", '{0:.2f}'.format(finish-start), 'seconds')
    return sp3matched

# IONOSPHERE MODEL INTERPOLATION
def ionosphere_interp(station, unit="meter", system="G", band="L1", epoch_list=None):
    """
    This function interpolates the TEC value from NASA Global Ionosphere
    Model at station location. The unit of the output can be tecu or meters.
    If function externally used, satellite system (G/R/E/C/J/I/S) and 
    frequency band must be specified manually.
    """
    ionosphereFile = ionFileName(station.epoch)
    tec = readFile.read_ionFile(ionosphereFile)

    obsList = _observation_picker_by_band(station, system, band) # observables

    lat, lon, ellHeight = cart2ell(station.approx_position[0], station.approx_position[1], station.approx_position[2])

    geocentric_lat = geocentric_latitude(lat)
    geocentric_lon = lon

    if type(epoch_list)==type(None):
        observationEpochList = station.observation.epoch
    else:
        observationEpochList = epoch_list

    Latitude  = _np.linspace(87.5,-87.5,71)
    Longitude = _np.linspace(-180,180,72)

    for i in range(len(Latitude)):
        if Latitude[i] > geocentric_lat > Latitude[i+1]:
            IndexLat = i
            break

    for i in range(len(Longitude)):
        if Longitude[i] < geocentric_lon < Longitude[i+1]:
            IndexLon = i
            break

    dlat = 2.5
    dlon = 5.0
    p = (geocentric_lon-Longitude[IndexLon])/dlon
    q = (geocentric_lat-Latitude[IndexLat+1])/dlat
    E = (1-p)*(1-q)*tec[:,IndexLat+1,IndexLon]+p*(1-q)*tec[:,IndexLat+1,IndexLon+1]+q*(1-p)*tec[:,IndexLat,IndexLon]+p*q*tec[:,IndexLat,IndexLon+1]
    tecFinal = _np.zeros(len(observationEpochList))
    
    epochStart = _pd.Timestamp(station.epoch)
    j = 0
    for i in range(12):
        epochEnd = epochStart + _pd.Timedelta(hours=2)
        while observationEpochList[j] < epochEnd:
            tecFinal[j] = ((epochEnd-observationEpochList[j])/(epochEnd-epochStart))*E[i]+((observationEpochList[j]-epochStart)/(epochEnd-epochStart))*E[i+1]
            j +=1
            if j == len(observationEpochList):
                break
        epochStart += _pd.Timedelta(hours=2)

    if unit.lower()=="tecu":
        ionDelay = 0.1*tecFinal # unit 0.1 TECU to 1 TECU
    elif unit.lower() in {"meter", "meters", "m"}:
        ionDelay = (0.1*tecFinal*40.3*1e16)/obsList[4]**2
    return ionDelay



# ===========================================================
# ========================= imports =========================
import numpy as _np
# ===========================================================
__all__ = ["ell2cart", "cart2ell","ell2topo"]

class _Ellipsoid:
    def __init__(self,a,b):
        self.a = a
        self.b = b
        self.f = (a-b)/a # flattening: $f = \frac{a-b}{a}$
        self.e1 = _np.sqrt((a**2-b**2)/a**2) # first eccentricity  : $e   = \sqrt{\frac{a^2-b^2}{a^2}}$
        self.e2 = _np.sqrt((a**2-b**2)/b**2) # second eccentricity : $e^' = \sqrt{\frac{a^2-b^2}{b^2}}$
    def radiusOfCurvature(self,phi):
        # Meridian radius of curvature (M)
        M = self.a*(1-self.e1**2)/(1-self.e1**2*_np.sin(_np.deg2rad(phi))**2)**(3/2) #$M = \frac{a(1 - e^2)}{(1 - e^2\sin(\phi)^2)^(3/2)}$
        # Prime vertical radius of curvature (N)
        N = self.a/(1-self.e1**2*_np.sin(_np.deg2rad(phi))**2)**(1/2) #$N = \frac{a}{(1 - e^2\sin(\phi)^2)^(1/2)} $
        return M, N
# -----------------------------------------------------------------------------
def _ellipsoid(ellipsoidName):
    axes = {'GRS80'  : [6378137.000, 6356752.314140],
            'WGS84'  : [6378137.000, 6356752.314245],
            'Hayford': [6378388.000, 6356911.946000]}[ellipsoidName]
    a, b = axes[0], axes[1]
    return _Ellipsoid(a,b)

def ell2cart(lat, lon, h, ellipsoid = 'GRS80'):
    """
    This function converts geodetic coordinates to 3D cartesian coordinates
    """
    ellipsoid = _ellipsoid(ellipsoid)
    lat = lat * _np.pi / 180 # in radians
    lon = lon * _np.pi / 180 # in radians
    N = ellipsoid.a / _np.sqrt(1-ellipsoid.e1**2 * _np.sin(lat)**2)
    x = (N + h) * _np.cos(lat) * _np.cos(lon)
    y = (N + h) * _np.cos(lat) * _np.sin(lon)
    z = ((1-ellipsoid.e1**2) * N + h) * _np.sin(lat)
    return x,y,z

def cart2ell(x, y, z, ellipsoid = 'GRS80'):
    """
    This function converts 3D cartesian coordinates to geodetic coordinates
    """
    ellipsoid = _ellipsoid(ellipsoid) # create an ellipsoid instance
    lon = _np.arctan2(y,x) # $\lambda = \atan\frac{y}{x}$
    p = _np.sqrt(x**2+ y**2) # $p = \sqrt{x^2+y^2}$
    if p < 1e-3:
        # poles exception
        if z<0:
            lat = -_np.pi/2
            h = z + ellipsoid.b
        else:
            lat = _np.pi/2
            h = z - ellipsoid.b
        return _np.rad2deg(lat), _np.rad2deg(lon), h
    N_init = ellipsoid.a # initial value of prime vertical radius N
    h_init = _np.sqrt(x**2 + y**2 + z**2) - _np.sqrt(ellipsoid.a * ellipsoid.b)
    lat_init = _np.arctan2(z, (1 - N_init * ellipsoid.e1**2 / (N_init + h_init)) * p)
    while True:
        N = ellipsoid.a / _np.sqrt(1-(ellipsoid.e1**2 * _np.sin(lat_init)**2))
        h = (p / _np.cos(lat_init)) - N
        lat = _np.arctan2(z, (1 - N * ellipsoid.e1**2 / (N + h)) * p)
        if _np.abs(lat_init - lat) < 1e-8 and _np.abs(h_init - h) < 1e-8:
            break
        lat_init = lat
        h_init   = h
    return _np.rad2deg(lat), _np.rad2deg(lon), h

def cart2ell_direct(x, y, z, ellipsoid = 'GRS80'):
    """
    This function converts 3D cartesian coordinates to geodetic coordinates
    """
    ellipsoid = _ellipsoid(ellipsoid) 
    lon = _np.arctan2(y,x) 
    p = _np.sqrt(x**2+ y**2)
    beta = _np.arctan2(ellipsoid.a*z,ellipsoid.b*p)
    lat = _np.arctan2(z+ellipsoid.e2**2*ellipsoid.b*_np.sin(beta)**3,p-ellipsoid.e1**2*ellipsoid.a*_np.cos(beta)**3)
    N = ellipsoid.a / _np.sqrt(1-(ellipsoid.e1**2 *_np.sin(lat)**2))
    h = (p/_np.cos(lat))-N
    return _np.rad2deg(lat), _np.rad2deg(lon), h

def ell2topo(lat, lon, h):
    """
    Convert ellipsoidal coordinates to topocentric coordinates 
    """
    lat, lon = _np.deg2rad(lat), _np.deg2rad(lon) # convert degree to radian
    east = _np.matrix([[-_np.sin(lon)], [_np.cos(lon)], [0]])
    north = _np.matrix([[-_np.cos(lon)*_np.sin(lat)],
                        [-_np.sin(lon)*_np.sin(lat)],
                        [ _np.cos(lat)]])
    up = _np.matrix([[_np.cos(lon)*_np.cos(lat)],
                        [_np.sin(lon)*_np.cos(lat)],
                        [_np.sin(lat)]])
    return (east, north, up)

def geocentric_latitude(geodetic_latitude, ellipsoid = 'GRS80'):
    """ Converts geodetic latitude to geocentric latitude """
    ell = _ellipsoid(ellipsoid)
    return _np.rad2deg(_np.arctan(_np.tan(_np.deg2rad(geodetic_latitude))/(1-ell.f)**2))

def _distance_euclidean(x1, y1, z1, x2, y2, z2):
    """ Calculates the euclidean distance [m] using cartesian coordinates """
    return _np.sqrt((x1-x2)**2 + (y1-y2)**2 + (z1-z2)**2)




# ===========================================================
# ========================= imports =========================
import numpy as _np
# ===========================================================
def clock_interp(fit, interval):
    """ 
    This function is for the interpolation of SP3 clocks.
    Inputs:
        fit: 2nd degree polynomial coefficients from numpy polyfit function
             (numpy array)
        interval: seconds (integer)
    Output:
        interpClock: interpolated clocks (numpy array)
        epoch: 
    """
    epoch = _np.linspace(0, 86400 , int(86400/interval)+1) # 86400 seconds = 24 hours
    time = _np.array([[epoch[t]**deg for deg in range(2,-1,-1)] for t in range(int(86400/interval)+1)])
    interpClock = _np.array([[sum(fit*time[t,:])] for t in range(len(time))])
    return interpClock, epoch

def coord_interp(parameter, interval):
    """ 
    This function is for the interpolation of SP3 coordinates. It fits 
    16 degree polynomial to 4 hours (14400 seconds) period of SP3 Cartesian 
    coordinates (per 15 minutes) and returns to the interpolated coordinates
    given in desired interval. NOTE: For clock interpolation refer to 'clock_interp'
    Inputs:
        fit: 16th degree polynomial coefficients from numpy polyfit function
             (numpy array)
        interval: seconds (integer)
    Output:
        interpCoord: interpolated coordinates (numpy array)
    """
    epoch = _np.linspace(1800, 12600 , int(10800/interval)+1) # 3h validity interval within 4h
    time = _np.array([epoch**deg for deg in range(len(parameter)-1,-1,-1)])
    return _np.matmul(parameter,time)



"""
preprocess reads files and returns analysis-ready DataSet

gather_stations merges observations from sites according to specified pairing rules over the desired time intervals
"""
# ===========================================================
# ========================= imports =========================
import os
import glob
import numpy as np
import pandas as pd
import xarray as xr
import tempfile
import fnmatch
from pathlib import Path
from typing import Union, Literal, Any
from gnssvod.io.io import Observation
from gnssvod.io.readFile import read_obsFile
from gnssvod.io.exporters import export_as_nc
from gnssvod.position.interpolation import sp3_interp_fast
from gnssvod.position.position import gnssDataframe
from gnssvod.funcs.constants import _system_name
#-------------------------------------------------------------------------
#----------------- FILE SELECTION AND BATCH PROCESSING -------------------
#-------------------------------------------------------------------------
def preprocess(filepattern: dict,
               orbit: bool = True,
               interval: Union[str,pd.Timedelta,None] = None,
               keepvars: Union[list,None] = None,
               outputdir: Union[dict, None] = None,
               overwrite: bool = False,
               encoding: Union[None, Literal['default'], dict] = 'default',
               outputresult: bool = False,
               aux_path: Union[str, None] = None,
               approx_position: list[float] = None) -> dict[Any,list[Observation]]:
    """
    Reads and processes structured lists of RINEX observation files.
    
    Parameters
    ----------
    filepattern: dictionary 
        Dictionary of station names and UNIX-style patterns to match RINEX 
        observation files. For example filepattern={'station1':'/path/to/files/of/station1/*O'}
    
    orbit: bool (optional) 
        if orbit=True, will download orbit solutions and calculate Azimuth and Elevation parameters
        if orbit=False, will not calculate additional gnss parameters
        
    interval: string or None (optional)
        if interval = None, the observations will be returned at the same rate as they were saved
        if interval = pandas Timedelta or str, this will be used to resample (average) the obervations (e.g. interval="15s")
    
    keepvars: list of strings or None (optional)
        Defines what columns are kept after processing. This can help reduce the size of the saved data.
        For example keepvars = ['S1','S2','Azimuth','Elevation']
        If None, no columns are removed
        
    outputdir: None, dictionary (optional)
        A dictionary of station names and folders indicating where to save the preprocessed data
        For example outputdir={'station1':'/path/where/to/save/preprocessed/data'}
        Dictionary keys must be the same as in the filepattern argument
        Data will be saved as a netcdf file, recycling the original file name
        If this argument is None, data won't be saved

    encoding: None, str, dict (optional)
        This argument is used to control compression options when saving netCDF files.
        If None is passed, no variable encodings are used when saving the file.
        If string 'default' is passed, will save all SNR, VOD, Azimuth, and Elevation data with some default encoding.
        Default encoding will be {"dtype": "int16", "scale_factor": 0.1, "zlib": True, "_FillValue":-9999}
        If a dict is passed, it should contain per-variable encodings that are passed to xr.to_netcdf(). This enables
        finer customization of the encoding.

    overwrite: bool (optional)
        If False (default), RINEX files with an existing matching file in the 
        specified output directory will be skipped entirely

    outputresult: bool (optional)
        If True and outputdir is None, observation objects will also be returned as a dictionary.

    aux_path: string or None (optional)
        If orbit is true, some external auxilliary orbit and clock files will be required and automatically downloaded.
        aux_path sets the directory where these files should be downloaded (or where they may already be found).
        If None is passed (default), a temporary directory is created and cleaned up if the processing succeeds.

    approx_position: list (optional)
        Position of the antenna provided as a list of cartesian coordinates [X,Y,Z]. This argument can be used to replace the 
        approximate position taken from the source RINEX files. 
        It is mandatory if source RINEX files actually miss the "APPROX POSITION XYZ" information in the header and the 
        'orbit' option is True. 
        To convert geographic coordinates (lat, lon, h) to cartesian (X,Y,Z) use gnssvod.geodesy.coordinate.ell2cart(lat,lon,h).

    Returns
    -------
    If outputresult = True, returns a dictionary. There is one key per station name and each item contains the GNSS observation object read 
    from each input RINEX file. For example output={'station1':[gnssvod.io.io.Observation,gnssvod.io.io.Observation,...]}
    
    """
    # set up temporary directory if necessary
    if orbit and (aux_path is None):
        tmp_folder = tempfile.TemporaryDirectory()
        aux_path = tmp_folder.name
        print(f"Created a temporary directory at {aux_path}")
    else: 
        tmp_folder = None
    # grab all files matching the patterns
    filelist = get_filelist(filepattern)
    
    out = dict()
    for item in filelist.items():
        station_name = item[0]
        filelist = item[1]

        # checking which files will be skipped (if necessary)
        if (not overwrite) and (outputdir is not None):
            # gather all files that already exist in the outputdir
            files_to_skip = get_filelist({station_name:f"{outputdir[station_name]}*.nc"})
            files_to_skip = [os.path.basename(x) for x in files_to_skip[station_name]]
        else:
            files_to_skip = []
        
        # for each file
        result = []
        for i,filename in enumerate(filelist):
            # determine the name of the output file that will be saved at the end of the loop
            out_name = os.path.splitext(os.path.basename(filename))[0]+'.nc'
            # if the name of the saved output file is in the files to skip, skip processing
            if out_name in files_to_skip:
                print(f"{out_name} already exists, skipping.. (pass overwrite=True to overwrite)")
                continue # skip remainder of loop and go directly to next filename
            
            # read in the file
            x = read_obsFile(filename)
            print(f"Processing {len(x.observation):n} individual observations")

            # only keep required vars
            if keepvars is not None:
                x.observation = subset_vars(x.observation,keepvars)
                # update the observation_types list
                x.observation_types = x.observation.columns.to_list()

            # resample if required
            if interval is not None:
                x = resample_obs(x,interval)
                
            # calculate Azimuth and Elevation if required
            if orbit:
                # use a prescribed position if one was passed as argument
                if approx_position is not None:
                    x.approx_position = approx_position
                # check that an approximate position exists before proceeding
                if (x.approx_position == [0,0,0]) or (x.approx_position is None):
                    raise ValueError("Missing an approximate antenna position. Provide the argument 'approx_position' to preprocess()")
                print(f"Calculating Azimuth and Elevation")
                # note: orbit cannot be parallelized easily because it 
                # downloads and unzips third-party files in the current directory
                if not 'orbit_data' in locals():
                    # if there is no previous orbit data, the orbit data is returned as well
                    x, orbit_data = add_azi_ele(x, aux_path = aux_path)
                else:
                    # on following iterations the orbit data is tentatively recycled to reduce computational time
                    x, orbit_data = add_azi_ele(x, orbit_data, aux_path = aux_path)

            # make sure we drop any duplicates
            x.observation=x.observation[~x.observation.index.duplicated(keep='first')]
            
            # store result in memory if required
            if outputresult:
                result.append(x)
                
            # write to file if required
            if outputdir is not None:
                outpath = str(Path(outputdir[station_name],out_name))
                export_as_nc(ds = x.to_xarray(),
                             outpath = outpath,
                             encoding = encoding)
                print(f"Saved {len(x.observation):n} individual observations in {outpath}")

        # store station in memory if required
        if outputresult:
            out[station_name]=result

    # clean up temporary directory if one exists
    if tmp_folder is not None:
        tmp_folder.cleanup()
        print(f"Removed the temporary directory at {aux_path}")

    if outputresult:
        return out
    else:
        return

def subset_vars(df: pd.DataFrame,
                keepvars: list,
                force_epoch_system: bool = True) -> pd.DataFrame:
    # find all matches for all elements of keepvars
    keepvars = np.concatenate([fnmatch.filter(df.columns.tolist(),x) for x in keepvars])
    # subselect those of the required columns that are present 
    tokeep = np.intersect1d(keepvars,df.columns.tolist())
    # + always keep 'epoch' and 'SYSTEM' as they are required for calculating azimuth and elevation
    if force_epoch_system:
        tokeep = np.unique(np.concatenate((keepvars,['epoch','SYSTEM'])))
    else:
        tokeep = np.unique(keepvars)
    # find columns not to keep
    todrop = np.setdiff1d(df.columns.tolist(),tokeep)
    # drop unneeded columns
    if len(todrop)>0:
        df = df.drop(columns=todrop)
    # drop rows for which all of the required vars are NA
    df = df.dropna(how='all')
    return df

def resample_obs(obs: Observation, interval: str) -> Observation:
    # list all variables except SYSTEM and epoch as these are handled differently
    subset = np.setdiff1d(obs.observation.columns.to_list(),['epoch','SYSTEM'])
    # resample those variables using temporal averaging
    obs.observation = obs.observation[subset].groupby([pd.Grouper(freq=interval, level='Epoch'),pd.Grouper(level='SV')]).mean()
    # restore SYSTEM and epoch
    obs.observation['epoch'] = obs.observation.index.get_level_values('Epoch')
    obs.observation['SYSTEM'] = _system_name(obs.observation.index.get_level_values("SV"))
    obs.interval = pd.Timedelta(interval).seconds
    return obs

def add_azi_ele(obs: Observation, 
                orbit_data: Union[pd.DataFrame,None] = None, 
                aux_path: Union[str,None] = None) -> tuple[Observation,pd.DataFrame]:
    start_time = min(obs.observation.index.get_level_values('Epoch'))
    end_time = max(obs.observation.index.get_level_values('Epoch'))
    
    if orbit_data is None:
        do = True
    elif (orbit_data.start_time<start_time) and (orbit_data.end_time>end_time) and (orbit_data.interval==obs.interval):
        # if the orbit for the day corresponding to the epoch and interval is the same as the one that was passed, just reuse it. This drastically reduces the number of times orbit files have to be read and interpolated.
        do = False
    else:
        do = True
    
    if do:
        # read (=usually download) orbit data
        orbit = sp3_interp_fast(start_time, end_time, interval=obs.interval, aux_path=aux_path)
        # prepare an orbit object as well
        orbit_data = orbit
        orbit_data.start_time = orbit.index.get_level_values('Epoch').min()
        orbit_data.end_time = orbit.index.get_level_values('Epoch').max()
        orbit_data.interval = obs.interval
    else:
        orbit = orbit_data
    
    # calculate the gnss parameters (including azimuth and elevation)
    gnssdf = gnssDataframe(obs,orbit,cut_off=-10)
    # add the gnss parameters to the observation dataframe
    obs.observation = obs.observation.join(gnssdf[['Azimuth','Elevation']])
    # drop variables 'epoch' and 'SYSTEM' as they are not needed anymore by gnssDataframe
    obs.observation = obs.observation.drop(columns=['epoch','SYSTEM'])
    # update the observation_types list
    obs.observation_types = obs.observation.columns.to_list()
    return obs, orbit_data

def get_filelist(filepatterns: dict) -> dict:
    if not isinstance(filepatterns,dict):
        raise Exception(f"Expected the input of get_filelist to be a dictionary, got a {type(filepatterns)} instead")
    filelists = dict()
    for item in filepatterns.items():
        station_name = item[0]
        search_pattern = item[1]
        flist = glob.glob(search_pattern)
        if len(flist)==0:
            print(f"Could not find any files matching the pattern {search_pattern}")
        filelists[station_name] = flist
    return filelists


#--------------------------------------------------------------------------
#----------------- PAIRING OBSERVATION FILES FROM SITES -------------------
#-------------------------------------------------------------------------- 

def gather_stations(filepattern: dict,
                    pairings: dict,
                    timeintervals: Union[pd.IntervalIndex,None] = None,
                    keepvars: Union[list,None] = None,
                    outputdir: Union[dict, None] = None,
                    encoding: Union[None, Literal['default'], dict] = None,
                    outputresult: bool = False) -> dict[Any,pd.DataFrame]:
    """
    Merges observations from different sites according to specified pairing rules. The new dataframe will contain 
    a new index level corresponding to each site, with keys corresponding to station names.
    
    Parameters
    ----------
    filepattern: dictionary 
        Dictionary of any number station names and UNIX-style patterns to find the preprocessed NetCDF files 
        observation files. For example filepattern={'station1':'/path/to/files/of/station1/*.nc',
                                                    'station2':'/path/to/files/of/station2/*.nc'}
    
    pairings: dictionary 
        Dictionary of case names associated to a tuple of station names indicating which stations to gather
        For example pairings={'case1':('station1','station2')}
        If data is to be saved, the case name will be taken as filename.
        
    timeintervals: None, pd.IntervalIndex (optional)
        The time interval(s) over which to sequentially gather data.
        For example timeperiod=pd.interval_range(start=pd.Timestamp('1/1/2018'), periods=8, freq='D') will gather 
        data for each of the 8 days in timeperiod. Adequate sequential processing will avoid that the script attempts
        to load and pair too much data at once.
        If outputdir is not None, the frequency also defines how data is saved (here as daily files).
        If None, all files that can be found will be used.
        
    keepvars: list of strings or None (optional)
        Defines what columns are kept after the gathering is made. This helps reduce the size of the data when saved.
        For example keepvars = ['S1','S2','Azimuth','Elevation']
        If None, no columns are removed
        
    outputdir: dictionary (optional)
        A dictionary of station names and folders indicating where to save the gathered data
        For example outputdir={'case1':'/path/where/to/save/data'}
        Data will be saved as a netcdf file, the dictionary has to be consistent with the 'pairings' argument
        If this argument is None, data will not be saved

    encoding: None, str, dict (optional)
        This argument is used to control compression options when saving netCDF files.
        If None is passed, no variable encodings are used when saving the file.
        If string 'default' is passed, will save all SNR, Azimuth, and Elevation data with some default encoding.
        Default encoding will be {"dtype": "int16", "scale_factor": 0.1, "zlib": True, "_FillValue":-9999}
        If a dict is passed, it should contain per-variable encodings that are passed to xr.to_netcdf(). This enables
        finer customization of the encoding.

    outputresult: bool (optional)
        If True, observation objects will also be returned as a dictionary
        
    Returns
    -------
    If outputresult = True, returns a dictionary. There is one key per case, and the corresponding item is a
    pd.Dataframe containing the paired data.
    
    """
    # get all files for all stations
    filenames = get_filelist(filepattern)
    print(f'Extracting Epochs from files')
    # extract only Epoch timestamps from all files (should be fast enough)
    epochs = {key:[xr.open_mfdataset(x)['Epoch'].values for x in items] for key,items in filenames.items()}
    # get min and max timestamp for each file (will be used to select which files to read later)
    epochs_min = {key:[np.min(x) for x in items] for key,items in epochs.items()}
    epochs_max = {key:[np.max(x) for x in items] for key,items in epochs.items()}
    
    result=dict()
    for case_name, station_names in pairings.items():
        out = []
        print(f'----- Processing {case_name}')
        if timeintervals is None:
            timeintervals = pd.interval_range(start=epochs_min, end=epochs_max)
        for interval in timeintervals:
            print(f'-- Processing interval {interval}')
            iout = []
            # gather all data required for that interval
            for station_name in station_names:
                # check which files have data that overlaps with the desired time intervals
                isin = [interval.overlaps(pd.Interval(left=pd.Timestamp(tmin),
                    right=pd.Timestamp(tmax))) for tmin,tmax in zip(epochs_min[station_name],epochs_max[station_name])]
                print(f'Found {sum(isin)} file(s) for {station_name}')
                if sum(isin)>0:
                    print(f'Reading')
                    # open those files and convert them to pandas dataframes
                    idata = [xr.open_mfdataset(x).to_dataframe().dropna(how='all') \
                            for x in np.array(filenames[station_name])[isin]]
                    # concatenate
                    idata = pd.concat(idata)
                    # keep only data falling within the interval
                    idata = idata.loc[[x in interval for x in idata.index.get_level_values('Epoch')]]
                    # drop duplicates and sort the dataframes
                    idata = idata[~idata.index.duplicated()].sort_index(level=['Epoch','SV'])
                    # add the station data in the iout list
                    iout.append(idata)
                else:
                    iout.append(pd.DataFrame())
                    print(f"No data for station {station_name}.")
                    continue

            if not all([x.empty for x in iout]):
                print(f'Concatenating stations')
                iout = pd.concat(iout, keys=station_names, names=['Station'])

                # only keep required vars and drop potential empty rows
                if keepvars is not None:
                    iout = subset_vars(iout,keepvars,force_epoch_system=False)
                    if len(iout)==0:
                        print(f"No observations left after subsetting columns (argument 'keepvars')")
                        continue
                
                # output the data as .nc if required
                if outputdir:
                    ioutputdir = outputdir[case_name]
                    print(f'Saving result in {ioutputdir}')
                    # make destination path
                    ts = f"{interval.left.strftime('%Y%m%d%H%M%S')}_{interval.right.strftime('%Y%m%d%H%M%S')}"
                    filename = f"{case_name}_{ts}.nc"
                    outpath = str(Path(ioutputdir,filename))
                    # sort dimensions
                    ds = iout.to_xarray()
                    ds = ds.sortby(['Epoch','SV','Station'])
                    # write nc file
                    export_as_nc(ds = ds,
                        outpath = outpath,
                        encoding = encoding)
                    print(f"Saved {len(iout)} observations in {filename}")

                # add interval in memory if required
                if outputresult:
                    out.append(iout)
            else:
                print(f"No data at all for that interval, skipping..")

        # store case in memory if required
        if outputresult and len(out)>0:
            result[case_name] = pd.concat(out)

    return result


"""
Position computation and related functions
"""
# ===========================================================
# ========================= imports =========================
import time
from datetime import timedelta as _timedelta
import numpy as _np
import pandas as _pd
from operator import itemgetter as _itemgetter
from gnssvod.geodesy.coordinate import _distance_euclidean
from gnssvod.position.atmosphere import tropospheric_delay
from gnssvod.position.satellite import _reception_coord, _sagnac, _azel, _relativistic_clock
from gnssvod.funcs.constants import (_SYSTEM_RNX2, _SYSTEM_RNX3,
                                    _SYSTEM_NAME, _CLIGHT)
# ===========================================================
__all__ = ["spp","multipath"]

def spp(station, orbit, system="G", cut_off=7.0):
    start = time.time() # Time of start
    if len(system)>1:
        raise Warning("SPP does not support multiple satellite system | This feature will be implemented in the next version")
    observation_list = _observation_picker(station, system)
    gnss = gnssDataframe(station, orbit, system, cut_off)
    #-----------------------------------------------------------------------------
    if len(observation_list) >=2:
        carrierPhase1 = getattr(gnss,observation_list[0][2])
        carrierPhase2 = getattr(gnss,observation_list[1][2])
        pseudorange1  = getattr(gnss,observation_list[0][3])
        pseudorange2  = getattr(gnss,observation_list[1][3])
        frequency1 = observation_list[0][4]
        frequency2 = observation_list[1][4]
    else:
        raise Warning("Ionosphere-free combination is not available")
    # ----------------------------------------------------------------------------
    gnss["Ionosphere_Free"] = (frequency1**2*pseudorange1-frequency2**2*pseudorange2)/(frequency1**2-frequency2**2)
    gnss = gnss.dropna(subset = ['Ionosphere_Free'])
    gnss["Travel_time"] = gnss["Ionosphere_Free"] / _CLIGHT
    gnss["X_Reception"],gnss["Y_Reception"],gnss["Z_Reception"] = _reception_coord(gnss.X, gnss.Y, gnss.Z, gnss.Vx, gnss.Vy, gnss.Vz, gnss.Travel_time)
    epochList =gnss.index.get_level_values("Epoch").unique().sort_values()
    epoch_start = epochList[0] 
    epoch_offset= _timedelta(seconds=300) 
    epoch_interval = _timedelta(seconds=station.interval-0.000001)
    epoch_stop  = epochList[-1] + _timedelta(seconds=0.000001)
    approx_position = [station.approx_position[0], station.approx_position[1], station.approx_position[2]]
    receiver_clock = station.receiver_clock
    position_list = []
    while True:
        epoch_step = epoch_start + epoch_interval
        gnss_temp = gnss.xs((slice(epoch_start,epoch_step))).copy()
        for iter in range(6):
            distance = _distance_euclidean(approx_position[0],approx_position[1],approx_position[2], gnss_temp.X_Reception, gnss_temp.Y_Reception, gnss_temp.Z_Reception)
            gnss_temp["Distance"] = distance + _sagnac(approx_position[0],approx_position[1],approx_position[2], gnss_temp.X_Reception, gnss_temp.Y_Reception, gnss_temp.Z_Reception)
            gnss_temp["Azimuth"], gnss_temp["Elevation"], gnss_temp["Zenith"] = _azel(station.approx_position[0], station.approx_position[1], station.approx_position[2], gnss_temp.X, gnss_temp.Y, gnss_temp.Z, gnss_temp.Distance)
            gnss_temp["Tropo"] = tropospheric_delay(station.approx_position[0],station.approx_position[1],station.approx_position[2], gnss_temp.Elevation, station.epoch)
            coeffMatrix = _np.zeros([len(gnss_temp),4]) 
            coeffMatrix[:,0] = (approx_position[0] - gnss_temp.X_Reception) / gnss_temp.Distance
            coeffMatrix[:,1] = (approx_position[1] - gnss_temp.Y_Reception) / gnss_temp.Distance
            coeffMatrix[:,2] = (approx_position[2] - gnss_temp.Z_Reception) / gnss_temp.Distance
            coeffMatrix[:,3] =  1
            lMatrix = gnss_temp.Ionosphere_Free - gnss_temp.Distance + _CLIGHT * (gnss_temp.DeltaTSV + gnss_temp.Relativistic_clock - receiver_clock) - gnss_temp.Tropo
            lMatrix = _np.array(lMatrix)
            try:
                linearEquationSolution = _np.linalg.lstsq(coeffMatrix,lMatrix,rcond=None)
                xMatrix = linearEquationSolution[0]
                pos = [approx_position[0] + xMatrix[0], approx_position[1] + xMatrix[1], approx_position[2] + xMatrix[2], receiver_clock + xMatrix[3] / _CLIGHT]
                approx_position[0], approx_position[1] , approx_position[2], receiver_clock = pos[0], pos[1], pos[2], pos[3]
            except:
                print("Cannot solve normal equations for epoch", epoch_start,"| Skipping...")
        position_list.append(pos)
        epoch_start += epoch_offset
        epoch_step  += epoch_offset
        if (epoch_step - epoch_stop) > _timedelta(seconds=station.interval):
            break
    x_coordinate = _np.mean([pos[0] for pos in position_list])
    y_coordinate = _np.mean([pos[1] for pos in position_list])
    z_coordinate = _np.mean([pos[2] for pos in position_list])
    rec_clock    = _np.mean([pos[3] for pos in position_list])
    finish = time.time()     # Time of finish
    print("Pseudorange calculation is done in", "{0:.2f}".format(finish-start), "seconds.")
    return (x_coordinate, y_coordinate, z_coordinate, rec_clock)

def gnssDataframe(station, orbit, system="G+R+E+C+J+I+S", cut_off=7.0):
    try:
        system = _itemgetter(*system.split("+"))(_SYSTEM_NAME)
        if type(system)==str: system = tuple([system])
    except KeyError:
        raise Warning("Unknown Satellite System:", system, "OPTIONS: G-R-E-C-J-R-I-S")
    epochMatch = station.observation.index.intersection(orbit.index)
    gnss = _pd.concat([station.observation.loc[epochMatch].copy(), orbit.loc[epochMatch]], axis=1)
    gnss = gnss[gnss['SYSTEM'].isin(system)]
    gnss["Distance"] = _distance_euclidean(station.approx_position[0], station.approx_position[1], station.approx_position[2], gnss.X, gnss.Y, gnss.Z)
    gnss["Relativistic_clock"] = _relativistic_clock(gnss.X, gnss.Y, gnss.Z, gnss.Vx, gnss.Vy, gnss.Vz)
    gnss['Azimuth'], gnss['Elevation'], gnss['Zenith'] = _azel(station.approx_position[0], station.approx_position[1], station.approx_position[2], gnss.X, gnss.Y, gnss.Z, gnss.Distance)
    gnss = gnss.loc[gnss['Elevation'] > cut_off]
    gnss["Tropo"] = tropospheric_delay(station.approx_position[0],station.approx_position[1],station.approx_position[2], gnss.Elevation, station.epoch)
    return gnss

def multipath(station, system="G"):
    if len(system)>1:
        raise Warning("Multiple satellite system is not applicable for multipath | This feature will be implemented in next version.")
    observation_list = _observation_picker(station, system=system)
    observation = station.observation.dropna(subset=[observation_list[0][2],observation_list[1][2],observation_list[0][3],observation_list[1][3]])
    observation = observation.loc[observation.SYSTEM==_SYSTEM_NAME[system]].copy(deep=True)
    carrierPhase1 = getattr(observation,observation_list[0][2])
    carrierPhase2 = getattr(observation,observation_list[1][2])
    pseudorange1  = getattr(observation,observation_list[0][3])
    pseudorange2  = getattr(observation,observation_list[1][3])
    frequency1 = observation_list[0][4]
    frequency2 = observation_list[1][4]
    lam1 = _CLIGHT/frequency1 
    lam2 = _CLIGHT/frequency2 
    ioncoeff = (frequency1/frequency2)**2
    observation["Multipath1"] = pseudorange1 - (2/(ioncoeff-1)+1)*(carrierPhase1*lam1) + (2/(ioncoeff-1))*(carrierPhase2*lam2)
    observation["Multipath2"] = pseudorange2 - (2*ioncoeff/(ioncoeff-1))*(carrierPhase1*lam1) + (2*ioncoeff/(ioncoeff-1)-1)*(carrierPhase2*lam2)
    observation = observation.reorder_levels(['SV','Epoch'])
    observation = observation.sort_index()
    sv_list = observation.index.get_level_values('SV').unique()
    # ----------------------------------------------------------------------------
    Multipath1 = []
    Multipath2 = []
    for sv in sv_list:
        ObsSV = observation.loc[sv]
        multipathSV1 = []
        multipathSV2 = []
        j = 0
        for i in range(1, len(ObsSV)):
            if (ObsSV.iloc[i].epoch - ObsSV.iloc[i-1].epoch) > _pd.Timedelta('0 days 00:15:00'):
                multipath1 = ObsSV.iloc[j:i].Multipath1.values - _np.nanmean(ObsSV.iloc[j:i].Multipath1.values)
                multipath2 = ObsSV.iloc[j:i].Multipath1.values - _np.nanmean(ObsSV.iloc[j:i].Multipath1.values)
                multipathSV1.extend(multipath1)
                multipathSV2.extend(multipath2)
                j=i
        multipath1 = ObsSV.iloc[j:].Multipath1.values - _np.nanmean(ObsSV.iloc[j:].Multipath1.values)
        multipath2 = ObsSV.iloc[j:].Multipath1.values - _np.nanmean(ObsSV.iloc[j:].Multipath1.values)
        multipathSV1.extend(multipath1)
        multipathSV2.extend(multipath2)
        Multipath1.extend(multipathSV1)
        Multipath2.extend(multipathSV2)
    # Re-assign multipath values
    observation["Multipath1"] = Multipath1
    observation["Multipath2"] = Multipath2
    return observation

def _adjustment(coeffMatrix,LMatrix):
    NMatrix = _np.linalg.inv(_np.dot(_np.transpose(coeffMatrix), coeffMatrix))
    nMatrix = _np.matmul(_np.transpose(coeffMatrix), LMatrix)
    XMatrix = _np.dot(NMatrix, nMatrix)
    vMatrix = _np.dot(coeffMatrix, XMatrix) - LMatrix
    m0 = _np.sqrt(_np.dot(_np.transpose(vMatrix), vMatrix)/(len(LMatrix)-len(NMatrix)))
    diagN = _np.diag(NMatrix)
    rmse = m0*_np.sqrt(diagN)
    mp = _np.sqrt(rmse[0]**2+rmse[1]**2+rmse[2]**2)
    return XMatrix, rmse

def _observation_picker(station, system="G"):
    try:
        system = _SYSTEM_NAME[system.upper()]
    except KeyError:
        raise Warning("Unknown Satellite System:", system, "OPTIONS: G-R-E-C-J-R-I-S")
    #-------------------------------------------------------------------
    # RINEX-3
    if station.version.startswith("3"):
        observation_codes = station.observation.columns.tolist()
        system_observations = getattr(station.observation_types, system)
        band_list         = set("L" + code[1] for code in observation_codes if len(code)==3)
        channel_list      = set([code[2] for code in observation_codes if len(code)==3])
        obs_codes = []
        for band in band_list:
            if band in _SYSTEM_RNX3[system]:
                for channel in channel_list:
                    if (band+channel) in _SYSTEM_RNX3[system][band]["Carrierphase"] and (band+channel) in system_observations:
                        obs_codes.append([system,band,(band+channel),("C"+band[1]+channel),_SYSTEM_RNX3[system][band]["Frequency"]])
                        break
    # RINEX-2
    elif station.version.startswith("2"):
        observation_codes = station.observation.columns.tolist()
        system_observations = station.observation_types
        band_list         = set(code for code in observation_codes if code.startswith(("L")))
        obs_codes = []
        for band in band_list:
            if band in _SYSTEM_RNX2[system].keys():
                for code in _SYSTEM_RNX2[system][band]["Pseudorange"]:
                    if code in system_observations:
                        obs_codes.append([system,band,band,code,_SYSTEM_RNX2[system][band]["Frequency"]])
                        break
    obs_codes = sorted(obs_codes, key=_itemgetter(1))
    return (obs_codes[0],obs_codes[1])

def _observation_picker_by_band(station, system="G", band="L1"):
    #-------------------------------------------------------------------
    try:
        system = _SYSTEM_NAME[system.upper()]
        if band not in _SYSTEM_RNX3[system].keys():
            raise Warning(band,"band cannot be found in",system,"satellite system! Band options for",system,"system:",tuple(_SYSTEM_RNX3[system].keys()))
    except KeyError:
        raise Warning("Unknown Satellite System:", system, "OPTIONS: G-R-E-C-J-R-I-S")
    #-------------------------------------------------------------------
    
    # RINEX-3
    if station.version.startswith("3"):
        observation_codes = station.observation.columns.tolist()
        system_observations = getattr(station.observation_types, system)
        channel_list      = set([code[2] for code in observation_codes if len(code)==3])
        obs_codes = []
        if band in _SYSTEM_RNX3[system]:
            for channel in channel_list:
                if (band+channel) in _SYSTEM_RNX3[system][band]["Carrierphase"] and (band+channel) in system_observations:
                    obs_codes.append([system,band,(band+channel),("C"+band[1]+channel),_SYSTEM_RNX3[system][band]["Frequency"],("D"+band[1]+channel),("S"+band[1]+channel)])
                    break
    # RINEX-2
    elif station.version.startswith("2"):
        observation_codes = station.observation.columns.tolist()
        system_observations = station.observation_types
        obs_codes = []
        if band in _SYSTEM_RNX2[system].keys():
            for code in _SYSTEM_RNX2[system][band]["Pseudorange"]:
                if code in system_observations:
                    obs_codes.append([system,band,band,code,_SYSTEM_RNX2[system][band]["Frequency"],("D"+band[1]),("S"+band[1])])
                    break
    return (obs_codes[0])



"""
export_as_nc()
export_output()
"""
# ===========================================================
# ========================= imports =========================
import numpy as np
import xarray as xr
from fnmatch import fnmatch
from typing import Union, Literal
from pathlib import Path
# ===========================================================
def export_as_nc(ds: xr.Dataset, outpath: str, encoding: Union[None, Literal['default'], dict] = None):
    outdir = Path(outpath).parent
    outname = Path(outpath)
    # check that the output directory exists
    if not outdir.exists():
        outdir.mkdir()
    # delete file if it exists
    if outname.exists():
        outname.unlink()
    # save as NetCDF
    if encoding is not None:
        if isinstance(encoding, str):
            encoding = get_default_encodings(ds, encoding)
        ds.to_netcdf(outpath,encoding=encoding)
    else:
        ds.to_netcdf(outpath)

def get_default_encodings(ds: xr.Dataset, encoding: Literal['default']) -> dict:
    if encoding == 'default':
        enc_1decimal = {"dtype": "int16", "scale_factor": 0.1, "zlib": True, "_FillValue":-9999}
        to_compress_1decimal = [fnmatch(x,'S??') | 
                    fnmatch(x,'S?') | 
                    fnmatch(x,'Azimuth') | 
                    fnmatch(x,'Elevation') for x in list(ds.keys())]
        encoding_1decimal = {x:enc_1decimal for x in np.array(list(ds.keys()))[to_compress_1decimal]}
        enc_2decimal = {"dtype": "int16", "scale_factor": 0.2, "zlib": True, "_FillValue":-9999}
        to_compress_2decimal = [fnmatch(x,'VOD*') for x in list(ds.keys())]
        encoding_2decimal = {x:enc_2decimal for x in np.array(list(ds.keys()))[to_compress_2decimal]}
        encoding = encoding_1decimal | encoding_2decimal
    else:
        raise ValueError("Unexpected input value for encoding, expected 'default'")
    return encoding


"""
Class definitions for I/O opreations
"""
# ===========================================================
# ========================= imports =========================
import xarray as xr

# ======================================================================
class Observation:
    """
    Observations class for RINEX Observation (*.*o) files
    """
    def __init__(self, filename=None, epoch=None, observation=None, approx_position=None,
                 receiver_type=None, antenna_type=None, interval=None,
                 receiver_clock=None, version=None, observation_types=None):
        self.filename          = filename 
        self.epoch             = epoch 
        self.observation       = observation
        self.approx_position   = approx_position
        self.receiver_type     = receiver_type
        self.antenna_type      = antenna_type
        self.interval          = interval
        self.receiver_clock    = receiver_clock
        self.version           = version
        self.observation_types = observation_types

    def to_xarray(self) -> xr.Dataset:
        ds = self.observation.to_xarray()
        ds = ds.assign_attrs({'filename' : self.filename,
                            'observation_types' : self.observation_types,
                            'epoch' : self.epoch.isoformat(),
                            'approx_position' : self.approx_position})
        return ds

class _ObservationTypes:
    def __init__(self, ToB_GPS=None, ToB_GLONASS=None, ToB_GALILEO=None,
                 ToB_COMPASS=None, ToB_QZSS=None, ToB_IRSS=None, ToB_SBAS=None):
        self.GPS     = ToB_GPS
        self.GLONASS = ToB_GLONASS
        self.GALILEO = ToB_GALILEO
        self.COMPASS = ToB_COMPASS
        self.QZSS    = ToB_QZSS
        self.IRSS    = ToB_IRSS
        self.SBAS    = ToB_SBAS
# ======================================================================

# ======================================================================
class Header:
    """
    Header class for RINEX Observation (*.*o) files
    """
    def __init__(self, filename=None, approx_position=None, receiver_type=None, 
                 antenna_type=None, start_date=None, end_date=None,
                 version=None, observation_types=None):
        self.filename          = filename
        self.approx_position   = approx_position
        self.receiver_type     = receiver_type
        self.antenna_type      = antenna_type
        self.start_date        = start_date
        self.end_date          = end_date
        self.version           = version
        self.observation_types = observation_types
# ======================================================================

# ======================================================================
class Navigation:
    """
    Navigation class for RINEX Observation (*.*n/p) files
    """
    def __init__(self, epoch = None, navigation = None, version = None):
        self.epoch           = epoch
        self.navigation      = navigation
        self.version         = version
# ======================================================================

# ======================================================================
class PEphemeris:
    """
    Class definition for SP3 file (Precise Ephemeris)
    """
    def __init__(self, epoch=None, ephemeris=None):
        self.epoch = epoch
        self.ephemeris = ephemeris



"""
hemibuild is used to build an equi-angular hemispheric grid
"""
# ===========================================================
# ========================= imports =========================
import os
import time
import glob
import datetime
import numpy as np
import pandas as pd
import xarray as xr
import multiprocessing
import pdb
from matplotlib.patches import Rectangle, Circle
# ===========================================================
"""
Class definition for hemispheric polar grid object
"""
# ======================================================================
class Hemi:
    """
    Hemispheric grid class
    """
    def __init__(self,angular_resolution,grid,elelims,azilims,CellIDs):
        self.angular_resolution = angular_resolution
        self.ncells = len(grid)
        self.grid = grid
        self.coords = self.grid.loc[:,['azi','ele']]
        self.elelims = elelims
        self.azilims = azilims
        self.CellIDs = CellIDs

    def patches(self):
        '''
        return a series of patches
        '''
        def plotpatch(dfrow):
            azimin = np.deg2rad(dfrow.azimin)
            elemax = 90-dfrow.elemax
            azimax = np.deg2rad(dfrow.azimax)
            elemin = 90-dfrow.elemin
            return(Rectangle([azimin,elemax],azimax-azimin,elemin-elemax,fill=True))
        patches = self.grid.apply(plotpatch,axis=1)
        return(patches.rename('Patches'))
    
    def add_CellID(self,
                  df: pd.DataFrame,
                  aziname: str='Azimuth',
                  elename: str='Elevation',
                  idname: str='CellID',
                  drop: bool=True):
        '''
        return the index of the grid cell where each observation belongs
        '''
        # check that columns specified by aziname and elename exist in df
        if not aziname in df:
            raise ValueError(f"No column '{aziname}' in the dataframe, indicate which column should be used with azi='ColumnName'")
        if not aziname in df:
            raise ValueError(f"No column '{elename}' in the dataframe, indicate which column should be used with ele='ColumnName'")

        # extract a subset of the df which we can manipulate 
        idf = df.loc[:,(aziname,elename)]
        # remove all data missing azi or ele
        idf = idf[~idf.isnull().any(axis=1)]
        # use modulo to ensure all azimuths are [0-360] (not i.e. -10 or 370)
        idf[aziname] = np.mod(idf[aziname]+360*10,360)
        # first cut the data by elevation band
        idf['eleind'] = pd.cut(idf[elename],
                              bins = np.concatenate((np.flip(self.elelims),[90])),
                              labels = np.flip(list(range(len(self.elelims)))))
        
        # define a function that retrieves indices for each elevation band, cutting with azimuthal edges specific to the given band
        def azicut(idf):
            # if elevation band contains no obs, just return empty result
            if len(idf)==0:
                idf[idname] = pd.Series(dtype=int)
                return(idf[idname])
            # find out which band we are in
            iele = idf['eleind'].iloc[0]
            # retrieve the corresponding azimuthal edges
            iazilims = self.azilims[iele]
            # cut data with azimuthal edges to retrieve azimuthal index
            idf['aziind'] = pd.cut(idf[aziname], 
                                  bins = np.concatenate((iazilims,[360])),
                                  labels = False,
                                  right = False)
            # return the corresponding CellID values
            idf[idname] = self.CellIDs[iele][idf['aziind'].values]
            return(idf[idname])

        # apply the azicut function to each elevation band
        idf=idf.groupby('eleind',group_keys=False, observed=False)[['eleind',aziname]].apply(azicut) # groupby will drop rows with eleind=NaN
        if idname in df:
            df = df.drop(columns=idname)
        # if drop is True, we are returning the input df with only rows that have a CellID
        
        if drop:
            return(df.join(idf,how='inner'))
        # if drop is False, we are returning the entire input df, including NaN CellIDs
        else:
            return(df.join(idf,how='left'))
        
    
    # plot(), plot empty grid, or if passing dataframe + ID name + var name, make a join and plot the data

#-------------------------------------------------------------------------
#----------------- building hemispheric grids and meshes -------------------
#-------------------------------------------------------------------------
def hemibuild(angular_resolution,cutoff=0):
    """
    Calculates a hemispheric grid where cells have approximately equal angular size. Returns grid properties as dataframe.
    
    Parameters
    ----------
    angular_resolution: numeric
        defines the diameter of the central (zenith) cell in degrees. 
        Rings of surrounding cells are build until the cutoff elevation angle is met
    
    cutoff: numeric (default 0)
        defines the elevation angle (in degrees) at which the hemispheric grid stops
        
    Returns
    -------
    dataframe of cell IDs with edge and center coordinates

    References
    ----------
    Beckers, B., & Beckers, P. (2012). A general rule for disk and hemisphere partition into equal-area cells. Computational Geometry, 45(7), 275-283.
    
    """
    # calculate number of rings
    ringlims = np.arange(angular_resolution/2,90-cutoff,angular_resolution)
    # calculate area of a cell
    cell_area = 2*np.pi*(1-np.cos(np.deg2rad(angular_resolution/2)))

    # instantiate empty lists
    cells = []
    elelims = []
    azilims = []
    CellIDs = []
    # add first zenith cell as df
    cells.append(pd.DataFrame(data={'azi':[0],
                                    'ele':[90],
                                    'azimin':[0],
                                    'azimax':[360],
                                    'elemin':[90-angular_resolution/2],
                                    'elemax':[90]}))
    elelims.append(90-angular_resolution/2)
    azilims.append(np.array([0]))
    CellIDs.append(np.array([0]))
    nextCellID = 1
    
    # add cells, ring by ring
    for iring, outer_radius in enumerate(ringlims[1:]):
        inner_radius = ringlims[iring]
        # calculate area of ring
        ring_area = 2*np.pi*(1-np.cos(np.deg2rad(outer_radius)))-2*np.pi*(1-np.cos(np.deg2rad(inner_radius)))
        # evenly split ring according to requested cell area
        numcells = round(ring_area/cell_area)
        # span of a single cell
        azispan = 360/numcells
        # generate CellIDs
        CellID = list(range(nextCellID,nextCellID+numcells))
        # also prepare the starting CellID for the next iteration
        nextCellID = CellID[-1]+1
        # add all cells into a list of dataframes
        azimin = np.linspace(0,360.0-azispan,numcells)
        cells.append(pd.DataFrame(data={'azi':np.linspace(azispan/2,360-azispan/2,numcells),
                                    'ele':np.full(numcells,90-(inner_radius+angular_resolution/2)),
                                    'azimin':azimin,
                                    'azimax':np.concatenate((azimin[1:],[360.0])),
                                    'elemin':np.full(numcells,90-outer_radius),
                                    'elemax':np.full(numcells,90-inner_radius)}))
        elelims.append(90-outer_radius)
        azilims.append(np.array(azimin))
        CellIDs.append(np.array(CellID))
    # concatenate all cells from all rings
    cells = pd.concat(cells).reset_index(drop=True).rename_axis('CellID')
    
    # instantiate Hemi object
    return Hemi(angular_resolution,cells,np.array(elelims),azilims,CellIDs)


# ===========================================================
# ========================= imports =========================
import os
import pandas as _pd
# ===========================================================

global _FILEPATH
_FILEPATH = os.path.dirname(os.path.abspath(__file__))

def is_IGS(siteName):
    igs = _pd.read_table(_FILEPATH+"/IGSList.txt",sep="\t", names = ["CODE","SITE","COUNTRY","LATITUDE","LONGITUDE","HEIGHT"], index_col = "CODE")
    if siteName.upper() not in igs.index.tolist():
        return False
    return True

def IGS(siteName):
    igs = _pd.read_table(_FILEPATH+"/IGSList.txt",sep="\t", names = ["CODE","SITE","COUNTRY","LATITUDE","LONGITUDE","HEIGHT"], index_col = "CODE")
    if is_IGS(siteName) == True:
        siteInfo = igs[igs.index==siteName.upper()]
        return siteInfo




# ===========================================================
# ========================= imports =========================
import os
from pathlib import Path
from gnssvod.funcs.date import doy2date
from gnssvod import download
from gnssvod.doc.IGS import is_IGS
from hatanaka import decompress_on_disk
# ===========================================================


def isfloat(value):
    """ To check if any variable can be converted to float or not """
    try:
        float(value)
        return True
    except ValueError:
        return False

def isint(value):
    """ To check if any variable can be converted to integer """
    try:
        int(value)
        return True
    except ValueError:
        return False
    
def iszip(fileName):
    if fileName.lower().endswith((".z",".zip",".gz")):
        return True
    else:
        return False

def does_a_zip_exist(fileName):
    if os.path.exists(fileName + ".z"):
        return fileName+".z"
    if os.path.exists(fileName + ".Z"):
        return fileName+".Z"
    if os.path.exists(fileName + ".gz"):
        return fileName+".gz"
    if os.path.exists(fileName + ".zip"):
        return fileName+".zip"
    else:
        return False

def isexist(fileName):
    if os.path.exists(fileName) == False:
        if not does_a_zip_exist(fileName):
            # --------- case where neither a file nor a zip of the file exist ------------
            print(f"This file does not exist: {fileName}")
            # --------- a download is attempted -------------------------
            extension = Path(fileName).suffix[1:].lower()
            if extension[-1] == "o":
                if is_IGS(fileName[:4]):
                    print(fileName + ".Z does not exist in working directory | Downloading...")
                    fileEpoch = doy2date(fileName)
                    download.get_rinex([fileName[:4]], fileEpoch, Datetime = True)
                    print(" | Download completed for", fileName + ".Z", " | Extracting...")
                    decompress_on_disk(fileName + ".Z", delete=True)
                else:
                    raise Warning(fileName,"does not exist in directory and cannot be found in IGS Station list!")
            elif extension == "rnx":
                if is_IGS(fileName[:4]):
                    print(fileName + " does not exist in working directory | Downloading...")
                    fileName = fileName.split(".")[0] + ".crx"
                    fileEpoch = doy2date(fileName)
                    download.get_rinex3([fileName[:4]], fileEpoch, Datetime = True)
                    print(" | Download completed for", fileName + ".gz", " | Extracting...")
                    decompress_on_disk(fileName + ".gz", delete=True)
                else:
                    raise Warning(fileName,"does not exist in directory and cannot be found in IGS Station list!")
            elif extension == "crx":
                if is_IGS(fileName[:4]):
                    print(fileName + ".gz does not exist in working directory | Downloading...")
                    fileEpoch = doy2date(fileName)
                    download.get_rinex3([fileName[:4]], fileEpoch, Datetime = True)
                    decompress_on_disk(fileName + ".gz", delete=True)
                else:
                    raise Warning(fileName,"does not exist in directory and cannot be found in IGS Station list!")
            elif extension[-1] in {"n","p","g"}:
                if is_IGS(fileName[:4]):
                    print(fileName + ".Z does not exist in working directory | Downloading...")
                    fileEpoch = doy2date(fileName)
                    download.get_navigation([fileName[:4]], fileEpoch, Datetime = True)
                    decompress_on_disk(fileName + ".gz", delete=True)
            elif extension in {"clk","clk_05s"}:
                download.get_clock(fileName)
            elif extension == "sp3":
                download.get_sp3(fileName)
            elif extension[-1].lower() == "i":
                download.get_ionosphere(fileName)
                decompress_on_disk(fileName + ".gz", delete=True)
            else:
                raise Warning("Unknown file extension:", extension)
                
# --------- case where a zip of the required file exists but not the file itself -------
        else:
            fileName = does_a_zip_exist(fileName)
            print(fileName + " exists | Extracting...")
            decompress_on_disk(fileName, delete=True)
            
 # --------- case where the required file and also a zip of it exist -------------------
    elif does_a_zip_exist(fileName):
        # one could decide to re-extract the zip again instead
        print(fileName + " exists | Reading...")
        
 # --------- case where the required file exists and is a zip file ---------------------
    elif iszip(fileName):
        import pdb
        decompress_on_disk(fileName, delete=True)
        print(fileName + " exists | Reading...") 
        
 # --------- case where the required file exists and is not a zip file -----------------
    else:
        print(fileName + " exists | Reading...")


"""
calc_vod calculates VOD according to specified pairing rules
"""
# ===========================================================
# ========================= imports =========================
import os
import time
import glob
import datetime
import numpy as np
import pandas as pd
import xarray as xr
import warnings
from gnssvod.io.preprocess import get_filelist
import pdb
#--------------------------------------------------------------------------
#----------------- CALCULATING VOD -------------------
#-------------------------------------------------------------------------- 

def calc_vod(filepattern,pairings,bands):
    """
    Combines a list of NetCDF files containing gathered GNSS receiver data, calculates VOD and returns that data.
    
    The gathered GNSS receiver data is typically generated with the function 'gather_stations'.
    
    VOD is calculated based on pairing rules referring to station names.
    
    Parameters
    ----------
    filepattern: dictionary 
        a UNIX-style pattern to find the processed NetCDF files.
        For example filepattern='/path/to/files/of/case1/*.nc'
    
    pairings: dictionary
        A dictionary of pairs of station names indicating first the reference station and second the ground station.
        For example pairings={'Laeg1':('Laeg2_Twr','Laeg1_Grnd')}

    bands: dictionary
        Dictionary of column names to be used for combining different bands
        For example bands={'VOD_L1':['S1','S1X','S1C']}
        
    Returns
    -------
    Dictionary of case names associated with dataframes containing the output for each case
    
    """
    files = get_filelist({'':filepattern})
    # read in all data
    data = [xr.open_mfdataset(x).to_dataframe().dropna(how='all') for x in files['']]
    # concatenate
    data = pd.concat(data)
    # calculate VOD based on pairings
    out = dict()
    for icase in pairings.items():
        iref = data.xs(icase[1][0],level='Station')
        igrn = data.xs(icase[1][1],level='Station')
        idat = iref.merge(igrn,on=['Epoch','SV'],suffixes=['_ref','_grn'])
        for ivod in bands.items():
            ivars = np.intersect1d(data.columns.to_list(),ivod[1])
            for ivar in ivars:
                irefname = f"{ivar}_ref"
                igrnname = f"{ivar}_grn"
                ielename = f"Elevation_grn"
                idat[ivar] = -np.log(np.power(10,(idat[igrnname]-idat[irefname])/10)) \
                            *np.cos(np.deg2rad(90-idat[ielename]))
            
            idat[ivod[0]] = np.nan
            for ivar in ivars:
                idat[ivod[0]] = idat[ivod[0]].fillna(idat[ivar])

        idat = idat[list(bands.keys())+['Azimuth_ref','Elevation_ref']].rename(columns={'Azimuth_ref':'Azimuth','Elevation_ref':'Elevation'})
        # store result in dictionary
        out[icase[0]]=idat
    return out


"""
Plotting module
"""
# ===========================================================
# ========================= imports =========================
import pandas as _pd
import numpy as _np
try:
    from mpl_toolkits.basemap import Basemap
    _BasemapStatus = True
except Exception as e:
    print("Warning: Basemap package cannot be imported |",e," error detected - Groundtrack plot disabled!")
    _BasemapStatus = False
import matplotlib.patches as _mpatches
from gnssvod.geodesy.coordinate import cart2ell_direct
from gnssvod.position.position import gnssDataframe, multipath, _observation_picker_by_band
from gnssvod.position.interpolation import ionosphere_interp
import matplotlib.pyplot as _plt
import matplotlib as _mptl
# ===========================================================

__all__ = ["skyplot","azelplot","bandplot","timelplot","groundtrack"]

# SKYPLOT
def skyplot(station, orbit, system = 'G', SVlist=None, color=None):
    # ----------------------------------------------------------------------
    if color != None:
        color = color.lower()
        if color not in {'ionosphere','snr1', 'snr2','snr3','snr4','snr5','snr6','snr7','snr8','snr9','multipath1', 'multipath2'}:
            raise Warning(color,"option for color mapping is not available for skyplot!")
    # ----------------------------------------------------------------------
    gnss = gnssDataframe(station, orbit, system)
    gnss = gnss.reorder_levels(["SV", "Epoch"])

    if color == 'multipath1' or color == 'multipath2':
        multipath_temp = multipath(station, system)
        epochMatch = gnss.index.intersection(multipath_temp.index)
        gnss = _pd.concat([gnss.loc[epochMatch].copy(), multipath_temp.loc[epochMatch].copy()], axis=1)
    
    # ----------------------------------------------------------------------
    SVList = gnss.index.get_level_values('SV').unique()
    gnss['svx'] = _np.radians(gnss.Azimuth.values)
    if SVlist == None:
        SVlist = SVList
    else:
        for sv in SVlist:
            if sv[0] not in {'G','R','C','E'} or sv[1:].isdigit() == False or len(sv)!=3:
                raise Warning("Invalid format: Please enter satellite(s) that you want to plot proper format. Exp. SVlist= ['G01', 'G02', 'G11',....] Program Stopped")
            elif sv not in SVList:
                SVlist.remove(sv)
                print('{} satellite not in SP3 file'.format(sv))
    if len(SVlist) == 0:
        raise Warning("Satellite(s) that you have entered not in SP3 file Program Stopped")
    fig = _plt.figure('Skyplot')
    figName = 'Skyplot'
    ax = fig.add_axes([0.1,0.1,0.8,0.8], polar=True)
    # -------------------------------------------------------------------------
    if color!=None:
        minValue, maxValue = 1e9, -1e9
        if color=="ionosphere":
            gnss['Ion_Delay'] = ionosphere_interp(station, unit  = 'meter', system= system, band = 'L1', epoch_list = gnss.epoch)
        for sv in SVlist:
            if color == 'ionosphere':
                if minValue > gnss.loc[sv].Ion_Delay.min():
                    minValue = gnss.loc[sv].Ion_Delay.min()
                if maxValue < gnss.loc[sv].Ion_Delay.max():
                    maxValue = gnss.loc[sv].Ion_Delay.max()
            elif color in {'snr1', 'snr2','snr3','snr4','snr5','snr6','snr7','snr8','snr9'}:
                obslist = _observation_picker_by_band(station, system, band = 'L'+color[-1])
                snr = getattr(gnss.loc[sv], obslist[6])
                if minValue > snr.min():
                    minValue = snr.min()
                if maxValue < snr.max():
                    maxValue = snr.max()
            elif color == 'multipath1':
                if minValue > gnss.loc[sv].Multipath1.min():
                    minValue = gnss.loc[sv].Multipath1.min()
                if maxValue < gnss.loc[sv].Multipath1.max():
                    maxValue = gnss.loc[sv].Multipath1.max()
            elif color == 'multipath2':
                if minValue > gnss.loc[sv].Multipath2.min():
                    minValue = gnss.loc[sv].Multipath2.min()
                if maxValue < gnss.loc[sv].Multipath2.max():
                    maxValue = gnss.loc[sv].Multipath2.max()
    # -------------------------------------------------------------------------
    for sv in SVlist:
        if color==None:
            skyplot=ax.scatter(gnss.loc[sv].svx, gnss.loc[sv].Zenith, s=3, cmap = _mptl.cm.jet)
            ax.text(gnss.loc[sv].svx[int(len(gnss.loc[sv].svx)/4)], gnss.loc[sv].Zenith[int(len(gnss.loc[sv].Zenith)/4)], sv, fontsize=9)
        elif color == 'ionosphere':
            skyplot=ax.scatter(gnss.loc[sv].svx, gnss.loc[sv].Zenith, s=3, c = gnss.loc[sv].Ion_Delay, cmap = _mptl.cm.jet, vmin=minValue, vmax=maxValue)
            ax.text(gnss.loc[sv].svx[int(len(gnss.loc[sv].svx)/4)], gnss.loc[sv].Zenith[int(len(gnss.loc[sv].Zenith)/4)], sv, fontsize=9)
        elif color in {'snr1', 'snr2','snr3','snr4','snr5','snr6','snr7','snr8','snr9'}:
            obslist = _observation_picker_by_band(station, system, band = 'L'+color[-1])
            snr = getattr(gnss.loc[sv], obslist[6])
            skyplot=ax.scatter(gnss.loc[sv].svx, gnss.loc[sv].Zenith, s=3, c = snr, cmap = _mptl.cm.jet, vmin=minValue, vmax=maxValue)
            ax.text(gnss.loc[sv].svx[int(len(gnss.loc[sv].svx)/4)], gnss.loc[sv].Zenith[int(len(gnss.loc[sv].Zenith)/4)], sv, fontsize=9)
        elif color == 'multipath1':
            skyplot=ax.scatter(gnss.loc[sv].svx, gnss.loc[sv].Zenith, s=3, c = gnss.loc[sv].Multipath1, cmap = _mptl.cm.jet, vmin=minValue, vmax=maxValue)
            ax.text(gnss.loc[sv].svx[int(len(gnss.loc[sv].svx)/4)], gnss.loc[sv].Zenith[int(len(gnss.loc[sv].Zenith)/4)], sv, fontsize=9)
        elif color == 'multipath2':
            skyplot=ax.scatter(gnss.loc[sv].svx, gnss.loc[sv].Zenith, s=3, c = gnss.loc[sv].Multipath2, cmap = _mptl.cm.jet, vmin=minValue, vmax=maxValue)
            ax.text(gnss.loc[sv].svx[int(len(gnss.loc[sv].svx)/4)], gnss.loc[sv].Zenith[int(len(gnss.loc[sv].Zenith)/4)], sv, fontsize=9)
    # Axes properties
    ax.set_theta_zero_location('N')
    ax.set_theta_direction(-1)
    ax.set_rmax(90.0)
    ax.set_yticks(range(0, 90, 10))
    ax.set_yticklabels(map(str, range(90, 0, -10)))
    if color in {'ionosphere','snr1', 'snr2','snr3','snr4','snr5','snr6','snr7','snr8','snr9','multipath1', 'multipath2'}:
        figName = figName + " ("+ color + ") for " + station.filename
        fig.colorbar(skyplot,extend="both")
    else:
        figName = figName + " for " + station.filename
    _plt.title(figName)
    fig.savefig(station.filename+"_Skyplot.png",transparent=True)

# AZEL PLOT
def azelplot(station, orbit, system = 'G', SVlist = None, color = None):
    if color != None:
        color = color.lower()
        if color not in {'ionosphere','snr1', 'snr2','snr3','snr4','snr5','snr6','snr7','snr8','snr9','troposphere'}:
            raise Warning(color,"option for color mapping is not available for azelplot!")
    # ----------------------------------------------------------------------
    gnss = gnssDataframe(station, orbit, system)
    gnss = gnss.reorder_levels(["SV", "Epoch"])

    if color == 'multipath1' or color == 'multipath2':
        multipath_temp = multipath(station, system)
        epochMatch = gnss.index.intersection(multipath_temp.index)
        gnss = _pd.concat([gnss.loc[epochMatch].copy(), multipath_temp.loc[epochMatch].copy()], axis=1)
    
    # ----------------------------------------------------------------------
    
    SVList = gnss.index.get_level_values('SV').unique()
    if SVlist == None:
        SVlist = SVList
    else:
        for sv in SVlist:
            if sv[0] not in {'G','R','C','E'} or sv[1:].isdigit() == False or len(sv)!=3:
                raise Warning('''Invalid format: Please enter satellite(s) that you want to plot proper format.
                Exp. SVlist= ['G01', 'G02', 'G11',....] 
                Program Stopped''')
            elif sv not in SVList:
                SVlist.remove(sv)
                print('{} satellite not in SP3 file'.format(sv))
    
    if len(SVlist) == 0:
        raise Warning("Satellite(s) not found in SP3 file | Program Stopped!")
    
    _plt.figure("Azimuth-Elevation Plot")
    figName = "Azimuth-Elevation Plot"
    _plt.xlabel('Azimuth', fontsize=12)
    _plt.ylabel('Elevation', fontsize=12)
    _plt.xlim(-180.0, 180.0)
    _plt.ylim(-1.0, 91.0)
    # -------------------------------------------------------------------------
    if color!=None:
        minValue, maxValue = 1e9, -1e9
        if color=="ionosphere":
            gnss['Ion_Delay'] = ionosphere_interp(station, unit  = 'meter', system= system, band = 'L1', epoch_list = gnss.epoch)
        for sv in SVlist:
            if color == 'ionosphere':
                if minValue > gnss.loc[sv].Ion_Delay.min():
                    minValue = gnss.loc[sv].Ion_Delay.min()
                if maxValue < gnss.loc[sv].Ion_Delay.max():
                    maxValue = gnss.loc[sv].Ion_Delay.max()
            elif color == 'troposphere':
                if minValue > gnss.loc[sv].Tropo.min():
                    minValue = gnss.loc[sv].Tropo.min()
                if maxValue < gnss.loc[sv].Tropo.max():
                    maxValue = gnss.loc[sv].Tropo.max()
            elif color in {'snr1', 'snr2','snr3','snr4','snr5','snr6','snr7','snr8','snr9'}:
                obslist = _observation_picker_by_band(station, system, band = 'L'+color[-1])
                snr = getattr(gnss.loc[sv], obslist[6])
                if minValue > snr.min():
                    minValue = snr.min()
                if maxValue < snr.max():
                    maxValue = snr.max()
            elif color == 'multipath1':
                if minValue > gnss.loc[sv].Multipath1.min():
                    minValue = gnss.loc[sv].Multipath1.min()
                if maxValue < gnss.loc[sv].Multipath1.max():
                    maxValue = gnss.loc[sv].Multipath1.max()
            elif color == 'multipath2':
                if minValue > gnss.loc[sv].Multipath2.min():
                    minValue = gnss.loc[sv].Multipath2.min()
                if maxValue < gnss.loc[sv].Multipath2.max():
                    maxValue = gnss.loc[sv].Multipath2.max()
    # -------------------------------------------------------------------------
    for sv in SVlist:
        if color == None:
            _plt.scatter(gnss.loc[sv].Azimuth, gnss.loc[sv].Elevation, s=3, cmap = _mptl.cm.jet, lw=0, vmin=minValue, vmax=maxValue)
            _plt.text(gnss.loc[sv].Azimuth[int(len(gnss.loc[sv].Azimuth)/3)], gnss.loc[sv].Elevation[int(len(gnss.loc[sv].Azimuth)/3)], sv, fontsize=9)
            _plt.text(gnss.loc[sv].Azimuth[int(len(gnss.loc[sv].Azimuth)/3*2)], gnss.loc[sv].Elevation[int(len(gnss.loc[sv].Azimuth)/3*2)], sv, fontsize=9)
        elif color == 'ionosphere':
            _plt.scatter(gnss.loc[sv].Azimuth, gnss.loc[sv].Elevation, s=3, c = gnss.loc[sv].Ion_Delay, cmap = _mptl.cm.jet, lw=0)
            _plt.text(gnss.loc[sv].Azimuth[int(len(gnss.loc[sv].Azimuth)/3)], gnss.loc[sv].Elevation[int(len(gnss.loc[sv].Azimuth)/3)], sv, fontsize=9)
            _plt.text(gnss.loc[sv].Azimuth[int(len(gnss.loc[sv].Azimuth)/3*2)], gnss.loc[sv].Elevation[int(len(gnss.loc[sv].Azimuth)/3*2)], sv, fontsize=9)
        elif color in {'snr1', 'snr2','snr3','snr4','snr5','snr6','snr7','snr8','snr9'}:
            obslist = _observation_picker_by_band(station, system, band = 'L'+color[-1])
            snr = getattr(gnss.loc[sv], obslist[6])
            _plt.scatter(gnss.loc[sv].Azimuth, gnss.loc[sv].Elevation, s=3, c = snr, cmap = _mptl.cm.jet, lw=0, vmin=minValue, vmax=maxValue)
            _plt.text(gnss.loc[sv].Azimuth[int(len(gnss.loc[sv].Azimuth)/3)], gnss.loc[sv].Elevation[int(len(gnss.loc[sv].Azimuth)/3)], sv, fontsize=9)
            _plt.text(gnss.loc[sv].Azimuth[int(len(gnss.loc[sv].Azimuth)/3*2)], gnss.loc[sv].Elevation[int(len(gnss.loc[sv].Azimuth)/3*2)], sv, fontsize=9)
        elif color == 'troposphere':
            _plt.scatter(gnss.loc[sv].Azimuth, gnss.loc[sv].Elevation, s=3, c = gnss.loc[sv].Tropo, cmap = _mptl.cm.jet, lw=0, vmin=minValue, vmax=maxValue)
            _plt.text(gnss.loc[sv].Azimuth[int(len(gnss.loc[sv].Azimuth)/3)], gnss.loc[sv].Elevation[int(len(gnss.loc[sv].Elevation)/3)], sv, fontsize=9)
            _plt.text(gnss.loc[sv].Azimuth[int(len(gnss.loc[sv].Azimuth)/3*2)], gnss.loc[sv].Elevation[int(len(gnss.loc[sv].Elevation)/3*2)], sv, fontsize=9)
    
    if color in {'ionosphere','snr1', 'snr2','snr3','snr4','snr5','snr6','snr7','snr8','snr9','troposphere'}: 
        _plt.colorbar(extend="both")
        figName = figName + " ("+ color + ") for " + station.filename
    else:
        figName = figName + " for " + station.filename
    _plt.title(figName)
    _plt.savefig(station.filename+"_AzelPlot.png",transparent=True)

# BANDPLOT
def bandplot(station, system = 'G', SVlist=None):
    gnss = station.observation
    # ----------------------------------------------------------------------
    SVList = gnss.index.get_level_values('SV').unique()

    if SVlist == None:
        SVlist = SVList
    else:
        for sv in SVlist:
            if sv[0] not in {'G','R','C','E'} or sv[1:].isdigit() == False or len(sv)!=3:
                raise Warning('''Invalid format: Please enter satellite(s) that you want to plot proper format.
                Exp. SVlist= ['G01', 'G02', 'G11',....] 
                Program Stopped''')
            elif sv not in SVList:
                SVlist.remove(sv)
                print('{} satellite not in SP3 file'.format(sv))
    
    if len(SVlist) == 0:
        raise Warning('''Satellite(s) that you have entered not in SP3 file
            Program Stopped''')
    RefPos = []
    line= 500
    for i in range(len(SVList)):
        RefPos.append(line+(i*500))
    
    RefPos = _pd.DataFrame(data = RefPos, index = SVList, columns = ['Visibility'])
    gnss = gnss.join(RefPos, how='inner')
    gnss = gnss.reorder_levels(['SV','Epoch'])
    gnss = gnss.sort_index()
    
    gnss['Time'] = gnss.epoch.dt.time
    _plt.figure("Band Plot")
    _plt.title("Band Plot", fontsize=12)
    _plt.xlabel('Time', fontsize=12)
    _plt.axes().get_yaxis().set_visible(False)
    
    for sv in SVlist:
        _plt.plot(gnss.loc[sv].Time, gnss.loc[sv].Visibility, linestyle='', marker='.',linewidth=1)
            #ax.xaxis.set_major_locator(mdates.HourLocator(byhour=range(24), interval=4))
        _plt.annotate(sv, xy=(gnss.loc[sv].Time[0], gnss.loc[sv].Visibility[0]+20), 
               size='medium', color='black')
    _plt.savefig(station.filename+"_BandPlot.png",transparent=True)

# TIME_ELEVATION PLOT
def timelplot(station, orbit, system = 'G', SVlist=None, color=None):
    if color != None:
        color = color.lower()
        if color not in {'ionosphere','troposphere','snr1', 'snr2','snr3','snr4','snr5','snr6','snr7','snr8','snr9','multipath1', 'multipath2'}:
            raise Warning(color,"option for color mapping is not available for timelplot!")
    # ----------------------------------------------------------------------
    gnss = gnssDataframe(station, orbit, system)
    gnss = gnss.reorder_levels(["SV", "Epoch"])

    if color == 'multipath1' or color == 'multipath2':
        multipath_temp = multipath(station, system)
        epochMatch = gnss.index.intersection(multipath_temp.index) # list of matched epochs and SVs
        gnss = _pd.concat([gnss.loc[epochMatch].copy(), multipath_temp.loc[epochMatch].copy()], axis=1)
    
    # ----------------------------------------------------------------------
    SVList = gnss.index.get_level_values('SV').unique()

    if SVlist == None:
        SVlist = SVList
    else:
        for sv in SVlist:
            if sv[0] not in {'G','R','C','E'} or sv[1:].isdigit() == False or len(sv)!=3:
                raise Warning('''Invalid format: Please enter satellite(s) that you want to plot proper format.
                Exp. SVlist= ['G01', 'G02', 'G11',....] 
                Program Stopped''')
            elif sv not in SVList:
                SVlist.remove(sv)
                print('{} satellite not in SP3 file'.format(sv))
    
    if len(SVlist) == 0:
        raise Warning('''Satellite(s) that you have entered not in SP3 file
            Program Stopped''')
    
    gnss['Time'] = gnss['epoch'].dt.time
    gnss['Seconds']= _pd.to_timedelta(gnss['epoch']).apply(lambda x: x.total_seconds())
    _plt.figure("Time-Elevation Plot")
    figName = "Time-Elevation Plot"
    _plt.xlabel('Time', fontsize=12)
    _plt.ylabel('Elevation', fontsize=12)
    _plt.ylim(-1.0, 91.0)
    _plt.grid()
    # -------------------------------------------------------------------------
    if color!=None:
        minValue, maxValue = 1e9, -1e9
        if color=="ionosphere":
            gnss['Ion_Delay'] = ionosphere_interp(station, unit  = 'meter', system= system, band = 'L1', epoch_list = gnss.epoch)
        for sv in SVlist:
            if color == 'ionosphere':
                if minValue > gnss.loc[sv].Ion_Delay.min():
                    minValue = gnss.loc[sv].Ion_Delay.min()
                if maxValue < gnss.loc[sv].Ion_Delay.max():
                    maxValue = gnss.loc[sv].Ion_Delay.max()
            elif color == 'troposphere':
                if minValue > gnss.loc[sv].Tropo.min():
                    minValue = gnss.loc[sv].Tropo.min()
                if maxValue < gnss.loc[sv].Tropo.max():
                    maxValue = gnss.loc[sv].Tropo.max()
            elif color in {'snr1', 'snr2','snr3','snr4','snr5','snr6','snr7','snr8','snr9'}:
                obslist = _observation_picker_by_band(station, system, band = 'L'+color[-1])
                snr = getattr(gnss.loc[sv], obslist[6])
                if minValue > snr.min():
                    minValue = snr.min()
                if maxValue < snr.max():
                    maxValue = snr.max()
            elif color == 'multipath1':
                if minValue > gnss.loc[sv].Multipath1.min():
                    minValue = gnss.loc[sv].Multipath1.min()
                if maxValue < gnss.loc[sv].Multipath1.max():
                    maxValue = gnss.loc[sv].Multipath1.max()
            elif color == 'multipath2':
                if minValue > gnss.loc[sv].Multipath2.min():
                    minValue = gnss.loc[sv].Multipath2.min()
                if maxValue < gnss.loc[sv].Multipath2.max():
                    maxValue = gnss.loc[sv].Multipath2.max()
    # -------------------------------------------------------------------------
    for sv in SVlist:
        if color == None:
            _plt.scatter(gnss.loc[sv].Seconds, gnss.loc[sv].Elevation, s=3, cmap = _mptl.cm.jet, lw=0, vmin=minValue, vmax=maxValue)
            _plt.text(gnss.loc[sv].Seconds[int(len(gnss.loc[sv].Seconds)/2)], gnss.loc[sv].Elevation[int(len(gnss.loc[sv].Elevation)/2)], sv, fontsize=9)
        elif color == 'troposphere':
            _plt.scatter(gnss.loc[sv].Seconds, gnss.loc[sv].Elevation, s=3, c = gnss.loc[sv].Tropo, cmap = _mptl.cm.jet, lw=0, vmin=minValue, vmax=maxValue)
            _plt.text(gnss.loc[sv].Seconds[int(len(gnss.loc[sv].Seconds)/2)], gnss.loc[sv].Elevation[int(len(gnss.loc[sv].Elevation)/2)], sv, fontsize=9)
        elif color in {'snr1', 'snr2','snr3','snr4','snr5','snr6','snr7','snr8','snr9'}:
            obslist = _observation_picker_by_band(station, system, band = 'L'+color[-1])
            snr = getattr(gnss.loc[sv], obslist[6])
            _plt.scatter(gnss.loc[sv].Seconds, gnss.loc[sv].Elevation, s=3, c = snr, cmap = _mptl.cm.jet, lw=0, vmin=minValue, vmax=maxValue)
            _plt.text(gnss.loc[sv].Seconds[int(len(gnss.loc[sv].Seconds)/2)], gnss.loc[sv].Elevation[int(len(gnss.loc[sv].Elevation)/2)], sv, fontsize=9)
        elif color == 'multipath1':
            _plt.scatter(gnss.loc[sv].Seconds, gnss.loc[sv].Elevation, s=3, c = gnss.loc[sv].Multipath1, cmap = _mptl.cm.jet, lw=0, vmin=minValue, vmax=maxValue)
            _plt.text(gnss.loc[sv].Seconds[int(len(gnss.loc[sv].Seconds)/2)], gnss.loc[sv].Elevation[int(len(gnss.loc[sv].Elevation)/2)], sv, fontsize=9)
        elif color == 'multipath2':
            _plt.scatter(gnss.loc[sv].Seconds, gnss.loc[sv].Elevation, s=3, c = gnss.loc[sv].Multipath2, cmap = _mptl.cm.jet, lw=0, vmin=minValue, vmax=maxValue)
            _plt.text(gnss.loc[sv].Seconds[int(len(gnss.loc[sv].Seconds)/2)], gnss.loc[sv].Elevation[int(len(gnss.loc[sv].Elevation)/2)], sv, fontsize=9)
        elif color == 'ionosphere':
            _plt.scatter(gnss.loc[sv].Seconds, gnss.loc[sv].Elevation, s=3, c = gnss.loc[sv].Ion_Delay, cmap = _mptl.cm.jet, lw=0, vmin=minValue, vmax=maxValue)
            _plt.text(gnss.loc[sv].Seconds[int(len(gnss.loc[sv].Seconds)/2)], gnss.loc[sv].Elevation[int(len(gnss.loc[sv].Elevation)/2)], sv, fontsize=9)
            
    if color in {"ionosphere","snr1","snr2","multipath1","multipath2","troposphere"}: 
        _plt.colorbar(extend="both")
        figName = figName + " ("+ color + ") for " + station.filename
    else:
        figName = figName + " for " + station.filename
    _plt.title(figName)
    _plt.savefig(station.filename+"_TimelPlot.png",transparent=True)

# GROUNDTRACK
if _BasemapStatus:
    def groundtrack(station, system = 'G', SVlist=None):
        gnss = station.observation
        SatEllCoor = cart2ell_direct(gnss.X.values, gnss.Y.values, gnss.Z.values)
        SVList = gnss.index.get_level_values('SVs').unique()
        gnss['Latitude'] = SatEllCoor[0]
        gnss['Longitude'] = SatEllCoor[1]
        gnss['EllHeight'] = SatEllCoor[2]
        gnss = gnss.reorder_levels(['SV','Epoch'])
    
        for sv in SVlist:
            if sv[0] not in {'G','R','C','E'} or sv[1:].isdigit() == False or len(sv)!=3:
                raise Warning('''Invalid format: Please enter satellite(s) that you want to plot proper format.
                Exp. SVlist= ['G01', 'G02', 'G11',....] 
                Program Stopped''')
            elif sv not in SVList:
                SVlist.remove(sv)
                print('{} satellite not in SP3 file'.format(sv))
        
        if len(SVlist) == 0:
            raise Warning('''Satellite(s) that you have entered not in SP3 file
                Program Stopped''')
        
        #------------------------start figure plot-------------------------------------#
        '''create map using BASEMAP
        llcrnrlat,llcrnrlon,urcrnrlat,urcrnrlon
        are the lat/lon values of the lower left and upper right corners
        of the map.
        lat_ts is the latitude of true scale.
        resolution = 'c' means use crude resolution coastlines'''
        
        m = Basemap(projection='merc',llcrnrlat=-80,urcrnrlat=80,\
                llcrnrlon=-180,urcrnrlon=180,lat_ts=20,resolution='c')
        m.drawcoastlines()
        m.fillcontinents(color='coral',lake_color='aqua')
        # draw parallels and meridians.
        m.drawparallels(_np.arange(-90.,91.,30.))
        m.drawmeridians(_np.arange(-180.,181.,60.))
        m.drawmapboundary(fill_color='aqua')
        
        coloropt = ['violet', 'red', 'wheat', 'yellow', 'orange', 'magenta', 'lime', 'ivory', 
                'gold', 'cyan', 'coral']
        
        for sv, i in zip(SVlist, range(len(SVlist))):
            lat = _pd.DataFrame()
            lon = _pd.DataFrame()
            lat[sv] = gnss.loc[sv].Latitude.values
            lon[sv] = gnss.loc[sv].Longitude.values
            lons, lats = m(lon[sv].values, lat[sv].values)
            m.scatter(lons, lats, color = coloropt[i], zorder=4, marker='.',s=3)
        recs = []
        for i in range(0,len(coloropt)):
            recs.append(_mpatches.Rectangle((0,0),1,1,fc=coloropt[i]))
        _plt.legend(recs, SVlist)
        _plt.title("Ground Track of GPS Satellites for {}".format(station.filename), fontweight='bold')
        _plt.savefig(station.filename+"_Groundtrack.png",transparent=True)



"""
obsFile function reads the observation file
navigationFile function reads the navigation file
sp3File function reads the SP3 file
clockFile function reads the clock file
"""
# ===========================================================
# ========================= imports =========================
import time
import datetime
import numpy as np
import pandas as pd
from gnssvod.funcs.checkif import (isfloat, isint, isexist)
from gnssvod.funcs.constants import _system_name
from gnssvod.io.io import Observation, Header, Navigation, _ObservationTypes
# ===========================================================

#-------------------------------------------------------------------------
#-------------------------------------------------------------------------
#----------------------------- NAVIGATION FILE ---------------------------
#-------------------------------------------------------------------------
#-------------------------------------------------------------------------
def read_navFile(navigationFile):
    """ Read navigation file """
    start = time.time()
    f = open(navigationFile)
    nav = f.readlines()
    line = 0
    while True:
        if 'RINEX VERSION / TYPE' in nav[line]:
            version = nav[line][0:-21].split()[0]
            line +=1
        elif 'END OF HEADER' in nav[line]:
            line +=1
            break
        else:
            line +=1
    del nav[0:line]
    nav = [lines.replace('E ','E0') for lines in nav]
    nav = [lines.replace('D','E') for lines in nav]
    nav = [lines.replace('0-','0 -') for lines in nav]
    nav = [lines.replace('1-','1 -') for lines in nav]
    nav = [lines.replace('2-','2 -') for lines in nav]
    nav = [lines.replace('3-','3 -') for lines in nav]
    nav = [lines.replace('4-','4 -') for lines in nav]
    nav = [lines.replace('5-','5 -') for lines in nav]
    nav = [lines.replace('6-','6 -') for lines in nav]
    nav = [lines.replace('7-','7 -') for lines in nav]
    nav = [lines.replace('8-','8 -') for lines in nav]
    nav = [lines.replace('9-','9 -') for lines in nav]
    nav = [lines.split() for lines in nav]
    ephemeris_list=[]
    svList = []
    PRN_old = "X00"
    epoch_old = datetime.datetime.now()
    while True:
        # --------------------------------------------------------------
        GPS = GLONASS = GALILEO = COMPASS = SBAS = False
        PRN = nav[0][0]
        if   "G" in PRN: GPS = True
        elif "R" in PRN: GLONASS = True
        elif "E" in PRN: GALILEO = True
        elif "C" in PRN: COMPASS = True
        elif "J" in PRN: QZSS = True
        elif "I" in PRN: IRSS = True
        elif "S" in PRN: SBAS = True
        else:
            if len(PRN)==1:
                PRN = 'G0' + PRN
            elif len(PRN)==2:
                PRN = 'G' + PRN
        # --------------------------------------------------------------

        if len(nav[0][1]) == 2:
            year = int(nav[0][1])
            if 79 < year < 100:
                year += 1900
            elif year <= 79:
                year += 2000
            else:
                raise Warning('Navigation year is not recognized! | Program stopped!')
        else:
            year = int(nav[0][1])
        month, day, hour, minute, second = int(nav[0][2]), int(nav[0][3]), int(nav[0][4]), int(nav[0][5]), int(nav[0][6][0])
        epoch = datetime.datetime(year   = year,
                                  month  = month,
                                  day    = day,
                                  hour   = hour,
                                  minute = minute,
                                  second = second)
        # --------------------------------------------------------------
        clockBias = nav[0][7] 
        relFeqBias = nav[0][8]
        transmissionTime = nav[0][9]
        if GLONASS or SBAS:
            x, vx, ax = float(nav[1][0]), float(nav[1][1]), float(nav[1][2])
            y, vy, ay = float(nav[2][0]), float(nav[2][1]), float(nav[2][2])
            z, vz, az = float(nav[3][0]), float(nav[3][1]), float(nav[3][2])
            health = float(nav[1][3])
            freqNumber = float(nav[2][3])
            operationDay = float(nav[3][3])
            roota, toe, m0, e, delta_n, smallomega, cus, cuc, crs, crc, cis, cic, idot, i0, bigomega0, bigomegadot = [np.nan for _ in range(16)]
            if PRN == PRN_old and epoch == epoch_old:
                ephemeris_list[-1] = [PRN, epoch, clockBias, relFeqBias, 
                                transmissionTime, roota, toe, m0, e, delta_n, 
                                smallomega, cus, cuc, crs, crc, cis, cic, 
                                idot, i0, bigomega0, bigomegadot,
                                x, y, z, vx, vy, vz, ax, ay, az,
                                health, freqNumber, operationDay]
            else:
                svList.append(PRN)
                ephemeris_list.append([PRN, epoch, clockBias, relFeqBias, 
                                    transmissionTime, roota, toe, m0, e, delta_n, 
                                    smallomega, cus, cuc, crs, crc, cis, cic, 
                                    idot, i0, bigomega0, bigomegadot,
                                    x, y, z, vx, vy, vz, ax, ay, az,
                                    health, freqNumber, operationDay])
            del nav[0:4]
        else:
            e = float(nav[2][1])
            m0 = float(nav[1][3])
            i0 = float(nav[4][0])
            toe = float(nav[3][0])
            cus = float(nav[2][2])
            cuc = float(nav[2][0])
            crs = float(nav[1][1])
            crc = float(nav[4][1])
            cis = float(nav[3][3])
            cic = float(nav[3][1])
            idot = float(nav[5][0])
            roota = float(nav[2][3])
            delta_n = float(nav[1][2])
            smallomega = float(nav[4][2])
            bigomega0 = float(nav[3][2])
            bigomegadot = float(nav[4][3])
            x, y, z, vx, vy, vz, ax, ay, az, health, freqNumber, operationDay = [np.nan for _ in range(12)]
            if PRN == PRN_old and epoch == epoch_old:
                ephemeris_list[-1] = [PRN, epoch, clockBias, relFeqBias, 
                                transmissionTime, roota, toe, m0, e, delta_n, 
                                smallomega, cus, cuc, crs, crc, cis, cic, 
                                idot, i0, bigomega0, bigomegadot,
                                x, y, z, vx, vy, vz, ax, ay, az,
                                health, freqNumber, operationDay]
            else:
                svList.append(PRN)
                ephemeris_list.append([PRN, epoch, clockBias, relFeqBias, 
                                    transmissionTime, roota, toe, m0, e, delta_n, 
                                    smallomega, cus, cuc, crs, crc, cis, cic, 
                                    idot, i0, bigomega0, bigomegadot,
                                    x, y, z, vx, vy, vz, ax, ay, az,
                                    health, freqNumber, operationDay])
            del nav[0:8]
        PRN_old = PRN
        epoch_old = epoch
        if len(nav) == 0:
            break
    columnNames = ["SV", "Epoch", "clockBias", "relFeqBias", "transmissionTime", "roota", "toe", "m0", "eccentricity", "delta_n", 
                "smallomega", "cus", "cuc", "crs", "crc", "cis", "cic", 
                "idot", "i0", "bigomega0", "bigomegadot",
                "x", "y", "z", "vx", "vy", "vz", "ax", "ay", "az",
                "health", "freqNumber", "operationDay"]
    ephemeris = pd.DataFrame(ephemeris_list, index=svList, columns=columnNames)
    ephemeris.index.name = 'SV'
    ephemeris["epoch"] = ephemeris.Epoch
    ephemeris.set_index('Epoch', append=True, inplace=True)
    ephemeris = ephemeris.reorder_levels(['Epoch', 'SV'])

    fileEpoch = datetime.date(year  = year,
                              month = month,
                              day   = day)
    f.close() # close the file
    finish = time.time()     # Time of finish
    print("Navigation file ", navigationFile," is read in", "{0:.2f}".format(finish-start), "seconds.")
    return Navigation(fileEpoch, ephemeris, version)

#-------------------------------------------------------------------------
#-------------------------------------------------------------------------
#----------------------------- OBSERVATION FILE --------------------------
#-------------------------------------------------------------------------
#-------------------------------------------------------------------------
def read_obsFile(observationFile,header=False):
    if observationFile.endswith(".Z")==True:
        raise Warning("All I/O functions take uncompressed files as an input (remove .Z/.gz from filename) | Next release will include this feature...")
    # check if observationFile exists or not
    isexist(observationFile)
    # open file
    f = open(observationFile, errors = 'ignore')
    obsLines = f.readlines()
    line = 0
    while True:
        if 'RINEX VERSION / TYPE' in obsLines[line]:
            version = obsLines[line][0:-20].split()[0]
            break
        else:
            line += 1
    f.close()
    if version.startswith("2"):
        return read_obsFile_v2(observationFile,header)
    elif version.startswith("3"):
        return read_obsFile_v3(observationFile,header)

def read_obsFile_v2(observationFile,header=False):
    """ Function that reads RINEX observation file """
    start = time.time()
    f = open(observationFile, errors = 'ignore')
    obsLines = f.readlines()

    line = 0
    ToB = []
    head = dict()
    while True:
        if 'RINEX VERSION / TYPE' in obsLines[line]:
            version = obsLines[line][0:-21].split()[0]
            line += 1
        elif 'REC # / TYPE / VERS' in obsLines[line]:
            receiver_type = obsLines[line][0:-20].split()
            line += 1
        elif 'ANT # / TYPE' in obsLines[line]:
            antenna_type = obsLines[line][0:-20].split()
            line += 1
        elif 'APPROX POSITION XYZ' in obsLines[line]:
            approx_position = obsLines[line][0:obsLines[line].index('A')].split()
            approx_position = [float(i) for i in approx_position]
            line += 1
        elif 'TYPES OF OBSERV' in obsLines[line]:
            ToB.extend(obsLines[line][0:obsLines[line].index('#')].split())
            obsNumber = int(ToB[0])
            line +=1
        elif 'TIME OF FIRST OBS' in obsLines[line]:
            start_date = obsLines[line][0:obsLines[line].index('TIME')].split()[0:-1]
            start_date = datetime.datetime(*np.array(start_date,dtype=float).round().astype(int))
            line += 1
        elif 'TIME OF LAST OBS' in obsLines[line]:
            end_date = obsLines[line][0:obsLines[line].index('TIME')].split()[0:-1]
            end_date = datetime.datetime(*np.array(end_date,dtype=float).round().astype(int))
            line += 1
        elif 'END OF HEADER' in obsLines[line]:
            line += 1
            break
        else:
            line += 1
    
    del obsLines[0:line]
    # --------------------------------------------------------------------------------------
    # if only the header was required, return a header object
    if header:
        f.close()
        finish = time.time()
        print("Observation file ", observationFile," is read in", "{0:.2f}".format(finish-start), "seconds.")
        return Header(f.name,approx_position,receiver_type,antenna_type,
                     start_date,end_date,version,ToB)
    # --------------------------------------------------------------------------------------
    obsLines = [lines.rstrip() for lines in obsLines] 
    obsList = []
    SVList= []
    epochList = []
    while True:
        # --------------------------------------------------------------------------------------
        while True:
            if 'COMMENT' in obsLines[0]:
                del obsLines[0]
                line += 1
            elif 'APPROX POSITION XYZ' in obsLines[0]:
                del obsLines[0]
                line += 1
            elif 'REC # / TYPE / VERS' in obsLines[0]:
                raise Warning("Receiver type is changed! | Exiting...")
            elif isint(obsLines[0][1:3])==False:
                print("Line", line, ":", obsLines[0]) # bu satırı sil!!!!
                del obsLines[0]
                line += 1
                print('Unexpected format between epochs! Line', line,'is deleted!')
            else:
                break
        #---------------------------------------------------------------------------------------
        year = int(obsLines[0][1:3])
        if 79 < year < 100:
            year += 1900
        elif year <= 79:
            year += 2000
        else:
            raise Warning('Observation year is not recognized! | Program stopped!')
        epoch = datetime.datetime(year = year, 
                                month =int(obsLines[0][4:6]),
                                day =int(obsLines[0][7:9]),
                                hour = int(obsLines[0][10:12]),
                                minute = int(obsLines[0][13:15]),
                                second  = int(obsLines[0][16:18])  if isint(obsLines[0][16:18])==True else 0)
                                #microsecond  = int(obsLines[0][20:27])  if isint(obsLines[0][19:20])==True else 0)
        epochList.append(epoch)
        eflag = int(obsLines[0][28:30])
        if eflag == 4:
            del obsLines[0]
            while True:
                if 'COMMENT' in obsLines[0]:
                    print(obsLines[0])
                    del obsLines[0]
                    line += 1
                else: 
                    break

        if len(obsLines[0]) == 80:
            receiver_clock = float(obsLines[0][-12:])
            obsLines[0] = obsLines[0][:-12]
        else:
            receiver_clock = 0
        #---------------------------------------------------------------------------------------
        NoSV  = int(obsLines[0][30:32])
        #---------------------------------------------------------------------------------------
        if len(obsLines[0][32:]) != 3*NoSV:
            noLine = [int(NoSV/12) if NoSV % 12 != 0 else int(NoSV/12)-1]
            for i in range(noLine[0]):
                obsLines[0] = obsLines[0] + obsLines[1][32:]
                del obsLines[1]
            if len(obsLines[0][32:]) != 3*NoSV:
                obsLines[0] = " " + obsLines[0].strip()
        #---------------------------------------------------------------------------------------
        SV=[obsLines[0][32:][i:i+3] for i in range(0, len(obsLines[0][32:]), 3)]
        SV = [i.replace('  ', 'G0') for i in SV]
        SV = [i.replace(' ', 'G') if i[0]==' ' else i.replace(' ', '0') for i in SV]
        SVList.extend(SV)
        #---------------------------------------------------------------------------------------
        obsEpoch = []
        del obsLines[0]
        rowNumber = np.ceil(obsNumber/5).astype('int')
        for i in range(0, rowNumber*NoSV, rowNumber): 
            for j in range(0, rowNumber):
                lineLenght = len(obsLines[i+j])
                if lineLenght != 80:
                    obsLines[i+j] += ' '*(80-lineLenght)
                if j!=0:
                    obsLines[i] += obsLines[i+j]
            obsLines[i] = [float(obsLines[i][j:j+14]) if isfloat(obsLines[i][j:j+14])==True else np.nan for j in range(0, 80*rowNumber, 16)]
            if len(obsLines[i]) != obsNumber:
                del obsLines[i][-(len(obsLines[i])-obsNumber):]
            obsLines[i].append(epoch)
            obsEpoch.append(obsLines[i])
        obsList.extend(obsEpoch)
        del obsLines[0:rowNumber*NoSV]
        if len(obsLines) == 0:
            break
    columnNames = ToB[1:]
    columnNames.append('Epoch')
    SVList = [x.replace(' ','0') if x[1]==' ' else x for x in SVList] # replace 'G 1' with 'G01' (etc)
    observation = pd.DataFrame(obsList, index=SVList, columns=columnNames)
    observation.index.name = 'SV'
    observation['epoch'] = observation.Epoch
    observation['Epoch'] = observation.Epoch
    observation.set_index('Epoch', append=True, inplace=True)
    observation = observation.reorder_levels(['Epoch', 'SV'])

    observation["SYSTEM"] = _system_name(observation.index.get_level_values("SV"))
    #---------------------------------------------------------------------------------------
    fileEpoch = datetime.date(year = epoch.year,
                                month = epoch.month,
                                day  = epoch.day)
    if len(epochList) == 1:
        interval = 30
    else:
        interval = epochList[1] - epochList[0]
        interval = interval.seconds
    f.close() # close the file
    #---------------------------------------------------------------------------------------
    finish = time.time()
    print("Observation file ", observationFile," is read in", "{0:.2f}".format(finish-start), "seconds.")
    return Observation(f.name, fileEpoch, observation, approx_position, receiver_type, antenna_type, interval, receiver_clock, version, ToB)

def read_obsFile_v3(obsFileName,header):
    start = time.time() # Time of start

    if obsFileName.endswith("crx"): obsFileName = obsFileName.split(".")[0] + ".rnx"
    f = open(obsFileName, errors = 'ignore') # open file
    obsLines = f.readlines() # read lines
    # =============================================================================

    line = 0
    ToB_GPS,ToB_GLONASS,ToB_GALILEO,ToB_COMPASS,ToB_QZSS,ToB_IRSS,ToB_SBAS = [], [], [], [], [], [], []
    while True:
        if 'RINEX VERSION / TYPE' in obsLines[line]:
            version = obsLines[line][0:-21].split()[0]
            line += 1
        elif 'REC # / TYPE / VERS' in obsLines[line]:
            receiver_type = obsLines[line][0:-20].split()
            line += 1
        elif 'ANT # / TYPE' in obsLines[line]:
            antenna_type = obsLines[line][0:-20].split()
            line += 1
        elif 'APPROX POSITION XYZ' in obsLines[line]:
            approx_position = obsLines[line][0:obsLines[line].index('A')].split()
            approx_position = [float(i) for i in approx_position]
            line += 1
        elif 'TIME OF FIRST OBS' in obsLines[line]:
            start_date = obsLines[line][0:obsLines[line].index('TIME')].split()[0:-1]
            start_date = datetime.datetime(*np.array(start_date,dtype=float).round().astype(int))
            line += 1
        elif 'TIME OF LAST OBS' in obsLines[line]:
            end_date = obsLines[line][0:obsLines[line].index('TIME')].split()[0:-1]
            end_date = datetime.datetime(*np.array(end_date,dtype=float).round().astype(int))
            line += 1
        elif 'SYS / # / OBS TYPES' in obsLines[line]:
            if obsLines[line][0] == "G": # GPS
                obsTypes = obsLines[line][0:obsLines[line].index('SYS')].split()
                obsNumber_GPS = int(obsTypes[1])
                ToB_GPS = obsTypes[2:]
                line += 1
                if obsNumber_GPS > 13:
                    for _ in range(int(np.ceil(obsNumber_GPS/13))-1):
                        ToB_GPS.extend(obsLines[line][0:obsLines[line].index('SYS')].split())
                        line += 1
            elif obsLines[line][0] == "R": # GLONASS
                obsTypes = obsLines[line][0:obsLines[line].index('SYS')].split()
                obsNumber_GLONASS = int(obsTypes[1])
                ToB_GLONASS = obsTypes[2:]
                line += 1
                if obsNumber_GLONASS > 13:
                    for _ in range(int(np.ceil(obsNumber_GLONASS/13))-1):
                        ToB_GLONASS.extend(obsLines[line][0:obsLines[line].index('SYS')].split())
                        line += 1
            elif obsLines[line][0] == "E": # GALILEO
                obsTypes = obsLines[line][0:obsLines[line].index('SYS')].split()
                obsNumber_GALILEO = int(obsTypes[1])
                ToB_GALILEO = obsTypes[2:]
                line += 1
                if obsNumber_GALILEO > 13:
                    for _ in range(int(np.ceil(obsNumber_GALILEO/13))-1):
                        ToB_GALILEO.extend(obsLines[line][0:obsLines[line].index('SYS')].split())
                        line += 1
            elif obsLines[line][0] == "C": # COMPASS
                obsTypes = obsLines[line][0:obsLines[line].index('SYS')].split()
                obsNumber_COMPASS = int(obsTypes[1])
                ToB_COMPASS = obsTypes[2:]
                line += 1
                if obsNumber_COMPASS > 13:
                    for _ in range(int(np.ceil(obsNumber_COMPASS/13))-1):
                        ToB_COMPASS.extend(obsLines[line][0:obsLines[line].index('SYS')].split())
                        line += 1
            elif obsLines[line][0] == "J": # QZSS
                obsTypes = obsLines[line][0:obsLines[line].index('SYS')].split()
                obsNumber_QZSS = int(obsTypes[1])
                ToB_QZSS = obsTypes[2:]
                line += 1
                if obsNumber_QZSS > 13:
                    for _ in range(int(np.ceil(obsNumber_QZSS/13))-1):
                        ToB_QZSS.extend(obsLines[line][0:obsLines[line].index('SYS')].split())
                        line += 1
            elif obsLines[line][0] == "I": # IRSS
                obsTypes = obsLines[line][0:obsLines[line].index('SYS')].split()
                obsNumber_IRSS = int(obsTypes[1])
                ToB_IRSS = obsTypes[2:]
                line += 1
                if obsNumber_IRSS > 13:
                    for _ in range(int(np.ceil(obsNumber_IRSS/13))-1):
                        ToB_IRSS.extend(obsLines[line][0:obsLines[line].index('SYS')].split())
                        line += 1
            elif obsLines[line][0] == "S": # SBAS
                obsTypes = obsLines[line][0:obsLines[line].index('SYS')].split()
                obsNumber_SBAS = int(obsTypes[1])
                ToB_SBAS = obsTypes[2:]
                line += 1
                if obsNumber_SBAS > 13:
                    for _ in range(int(np.ceil(obsNumber_SBAS/13))-1):
                        ToB_SBAS.extend(obsLines[line][0:obsLines[line].index('SYS')].split())
                        line += 1
        elif 'END OF HEADER' in obsLines[line]:
            line +=1
            break
        else:
            line +=1
    del obsLines[0:line]
    # =============================================================================
    # Type of Observations in RINEX File
    ToB = np.unique(np.concatenate((ToB_GPS,ToB_GLONASS,ToB_GALILEO,ToB_COMPASS,ToB_QZSS,ToB_IRSS,ToB_SBAS)))
    # Indices of each satellite constellation in ToB
    index_GPS = np.searchsorted(ToB,ToB_GPS)
    index_GLONASS = np.searchsorted(ToB,ToB_GLONASS)
    index_GALILEO = np.searchsorted(ToB,ToB_GALILEO)
    index_COMPASS = np.searchsorted(ToB,ToB_COMPASS)
    index_QZSS = np.searchsorted(ToB,ToB_QZSS)
    index_IRSS = np.searchsorted(ToB,ToB_IRSS)
    index_SBAS = np.searchsorted(ToB,ToB_SBAS)
    observation_types = _ObservationTypes(ToB_GPS,ToB_GLONASS,ToB_GALILEO,ToB_COMPASS,ToB_QZSS,ToB_IRSS,ToB_SBAS)
    # --------------------------------------------------------------------------------------
    # if only the header was required, return a header object
    if header:
        f.close()
        finish = time.time()
        print("Observation file ", obsFileName," is read in", "{0:.2f}".format(finish-start), "seconds.")
        return Header(f.name,approx_position,receiver_type,antenna_type,
                     start_date,end_date,version,ToB)
    # --------------------------------------------------------------------------------------
    # =============================================================================
    obsLines = [lines.rstrip() for lines in obsLines]
    obsList = []
    svList= []
    epochList = []
    currentline = 0
    last_update = start
    while True:
        if time.time()-last_update>30:
            # print an update on the progress every 30 seconds
            last_update = time.time()
            print(f'Processed {currentline} out of {len(obsLines)} lines ({currentline/len(obsLines)*100:.1f}%)')
        # =============================================================================
        while True:
            if 'COMMENT' in obsLines[currentline]:
                currentline += 1
            elif 'APPROX POSITION XYZ' in obsLines[currentline]:
                currentline += 1
            elif 'REC # / TYPE / VERS' in obsLines[currentline]:
                raise Warning("Receiver type is changed! | Exiting...")
            else:
                break
        # =============================================================================
        if obsLines[currentline][0] == ">":
            epochLine = obsLines[currentline][1:].split()
            if len(epochLine) == 8:
                epoch_year, epoch_month, epoch_day, epoch_hour, epoch_minute, epoch_second, epoch_flag, epoch_SVNumber = obsLines[currentline][1:].split()
                receiver_clock = 0 
            elif len(epochLine) == 9:
                epoch_year, epoch_month, epoch_day, epoch_hour, epoch_minute, epoch_second, epoch_flag, epoch_SVNumber, receiver_clock = obsLines[currentline][1:].split()
            else: 
                raise Warning("Unexpected epoch line format detected! | Program stopped!")
        else: 
            if currentline+1 == len(obsLines):
                # sometimes the last line is weird, especially if logging was interrupted
                print("The last line has an unexpected format and will be ignored")
                break # stop reading file
            else:
                print(f"\033[93mUnexpected format detected! | Tentatively skipping line #{currentline}")
                currentline += 1
                continue # skip remainder of loop
        # =========================================================================
        if epoch_flag in {"1","3","5","6"}:
            raise Warning("Deal with this later!")
        elif epoch_flag == "4":
            currentline += 1
            while True:
                if 'COMMENT' in obsLines[currentline]:
                    print(obsLines[currentline])
                    currentline += 1
                elif 'SYS / PHASE SHIFT' in obsLines[currentline]:
                    currentline +=1
                else: 
                    break
        else:
            # =========================================================================
            epoch = datetime.datetime(year = int(epoch_year), 
                                    month = int(epoch_month),
                                    day = int(epoch_day),
                                    hour = int(epoch_hour),
                                    minute = int(epoch_minute),
                                    second = int(float(epoch_second)))
            epochList.append(epoch)
            currentline += 1 # move beyond header line
            # =============================================================================
            epoch_SVNumber = int(epoch_SVNumber)
            for svLine in range(currentline,currentline+epoch_SVNumber):
                obsEpoch = np.full((1,len(ToB)), np.nan)
                svList.append(obsLines[svLine][:3])
                if obsLines[svLine].startswith("G"):
                    obsEpoch[0,index_GPS] = np.array([[float(obsLines[svLine][3:][i:i+14]) if isfloat(obsLines[svLine][3:][i:i+14])==True else np.nan for i in range(0,len(ToB_GPS)*16,16)]])
                elif obsLines[svLine].startswith("R"):
                    obsEpoch[0,index_GLONASS] = np.array([[float(obsLines[svLine][3:][i:i+14]) if isfloat(obsLines[svLine][3:][i:i+14])==True else np.nan for i in range(0,len(ToB_GLONASS)*16,16)]])
                    #obsEpoch[svLine,index_GLONASS]
                elif obsLines[svLine].startswith("E"):
                    obsEpoch[0,index_GALILEO] = np.array([[float(obsLines[svLine][3:][i:i+14]) if isfloat(obsLines[svLine][3:][i:i+14])==True else np.nan for i in range(0,len(ToB_GALILEO)*16,16)]])
                elif obsLines[svLine].startswith("C"):
                    obsEpoch[0,index_COMPASS] = np.array([[float(obsLines[svLine][3:][i:i+14]) if isfloat(obsLines[svLine][3:][i:i+14])==True else np.nan for i in range(0,len(ToB_COMPASS)*16,16)]])
                elif obsLines[svLine].startswith("J"):
                    obsEpoch[0,index_QZSS] = np.array([[float(obsLines[svLine][3:][i:i+14]) if isfloat(obsLines[svLine][3:][i:i+14])==True else np.nan for i in range(0,len(ToB_QZSS)*16,16)]])
                elif obsLines[svLine].startswith("I"):
                    obsEpoch[0,index_IRSS] = np.array([[float(obsLines[svLine][3:][i:i+14]) if isfloat(obsLines[svLine][3:][i:i+14])==True else np.nan for i in range(0,len(ToB_IRSS)*16,16)]])
                elif obsLines[svLine].startswith("S"):
                    obsEpoch[0,index_SBAS] = np.array([[float(obsLines[svLine][3:][i:i+14]) if isfloat(obsLines[svLine][3:][i:i+14])==True else np.nan for i in range(0,len(ToB_SBAS)*16,16)]])
                obsEpoch = np.append(obsEpoch,epoch)
                obsList.append(obsEpoch)
            # =============================================================================
            currentline += epoch_SVNumber # number of rows in epoch equals number of visible satellites in RINEX 3
        if currentline == len(obsLines):
            break
    """
    while True:
        # =============================================================================
        while True:
            if 'COMMENT' in obsLines[0]:
                del obsLines[0]
                line += 1
            elif 'APPROX POSITION XYZ' in obsLines[0]:
                del obsLines[0]
                line += 1
            elif 'REC # / TYPE / VERS' in obsLines[0]:
                raise Warning("Receiver type is changed! | Exiting...")
            else:
                break
        # =============================================================================
        if obsLines[0][0] == ">":
            epochLine = obsLines[0][1:].split()
            if len(epochLine) == 8:
                epoch_year, epoch_month, epoch_day, epoch_hour, epoch_minute, epoch_second, epoch_flag, epoch_SVNumber = obsLines[0][1:].split()
                receiver_clock = 0 
            elif len(epochLine) == 9:
                epoch_year, epoch_month, epoch_day, epoch_hour, epoch_minute, epoch_second, epoch_flag, epoch_SVNumber, receiver_clock = obsLines[0][1:].split()
            else: raise Warning("Unexpected epoch line format detected! | Program stopped!")
        else: raise Warning("Unexpected format detected! | Program stopped!")
        # =========================================================================
        if epoch_flag in {"1","3","5","6"}:
            raise Warning("Deal with this later!")
        elif epoch_flag == "4":
            del obsLines[0]
            while True:
                if 'COMMENT' in obsLines[0]:
                    print(obsLines[0])
                    del obsLines[0]
                    line += 1
                elif 'SYS / PHASE SHIFT' in obsLines[0]:
                    del obsLines[0]
                    #line += 1
                else: 
                    break
        else:
            # =========================================================================
            epoch = datetime.datetime(year = int(epoch_year), 
                                    month = int(epoch_month),
                                    day = int(epoch_day),
                                    hour = int(epoch_hour),
                                    minute = int(epoch_minute),
                                    second = int(float(epoch_second)))
            epochList.append(epoch)
            del obsLines[0] # delete epoch header line
            # =============================================================================
            epoch_SVNumber = int(epoch_SVNumber)
            for svLine in range(epoch_SVNumber):
                obsEpoch = np.full((1,len(ToB)), None)
                svList.append(obsLines[svLine][:3])
                if obsLines[svLine].startswith("G"):
                    obsEpoch[0,index_GPS] = np.array([[float(obsLines[svLine][3:][i:i+14]) if isfloat(obsLines[svLine][3:][i:i+14])==True else None for i in range(0,len(ToB_GPS)*16,16)]])
                elif obsLines[svLine].startswith("R"):
                    obsEpoch[0,index_GLONASS] = np.array([[float(obsLines[svLine][3:][i:i+14]) if isfloat(obsLines[svLine][3:][i:i+14])==True else None for i in range(0,len(ToB_GLONASS)*16,16)]])
                    #obsEpoch[svLine,index_GLONASS]
                elif obsLines[svLine].startswith("E"):
                    obsEpoch[0,index_GALILEO] = np.array([[float(obsLines[svLine][3:][i:i+14]) if isfloat(obsLines[svLine][3:][i:i+14])==True else None for i in range(0,len(ToB_GALILEO)*16,16)]])
                elif obsLines[svLine].startswith("C"):
                    obsEpoch[0,index_COMPASS] = np.array([[float(obsLines[svLine][3:][i:i+14]) if isfloat(obsLines[svLine][3:][i:i+14])==True else None for i in range(0,len(ToB_COMPASS)*16,16)]])
                elif obsLines[svLine].startswith("J"):
                    obsEpoch[0,index_QZSS] = np.array([[float(obsLines[svLine][3:][i:i+14]) if isfloat(obsLines[svLine][3:][i:i+14])==True else None for i in range(0,len(ToB_QZSS)*16,16)]])
                elif obsLines[svLine].startswith("I"):
                    obsEpoch[0,index_IRSS] = np.array([[float(obsLines[svLine][3:][i:i+14]) if isfloat(obsLines[svLine][3:][i:i+14])==True else None for i in range(0,len(ToB_IRSS)*16,16)]])
                elif obsLines[svLine].startswith("S"):
                    obsEpoch[0,index_SBAS] = np.array([[float(obsLines[svLine][3:][i:i+14]) if isfloat(obsLines[svLine][3:][i:i+14])==True else None for i in range(0,len(ToB_SBAS)*16,16)]])
                obsEpoch = np.append(obsEpoch,epoch)
                obsList.append(obsEpoch)
            # =============================================================================
            del obsLines[0:epoch_SVNumber] # number of rows in epoch equals number of visible satellites in RINEX 3
        if len(obsLines) == 0:
            break
    """
    # =============================================================================
    columnNames = ToB
    columnNames = np.append(ToB,'Epoch')
    svList = [x.replace(' ','0') if x[1]==' ' else x for x in svList] # replace 'G 1' with 'G01' (etc)
    obs = pd.DataFrame(obsList, index=svList, columns=columnNames)
    obs.index.name = 'SV'
    obs['epoch'] = obs.Epoch 
    obs['Epoch'] = obs.Epoch
    obs.set_index('Epoch', append=True, inplace=True)
    obs = obs.reorder_levels(['Epoch', 'SV'])
    obs["SYSTEM"] = _system_name(obs.index.get_level_values("SV"))
    # =============================================================================
    fileEpoch = datetime.date(year = epoch.year,
                                month = epoch.month,
                                day  = epoch.day)
    if len(epochList) == 1:
        interval = 30
    else:
        interval = epochList[1] - epochList[0]
        interval = interval.seconds
    f.close() # close the file
    # =============================================================================
    finish = time.time()     # Time of finish
    print("Observation file ", obsFileName," is read in", "{0:.2f}".format(finish-start), "seconds.")
    return Observation(f.name, fileEpoch, obs, approx_position, receiver_type, antenna_type, interval, receiver_clock, version, ToB)

#-------------------------------------------------------------------------
#-------------------------------------------------------------------------
#------------------------------- SP3 FILE --------------------------------
#-------------------------------------------------------------------------
#-------------------------------------------------------------------------
def read_sp3File(sp3file):
    start = time.time()
    isexist(sp3file)
    f = open(sp3file)
    sp3 = f.readlines()
    line = 0
    search = '/*'
    SVNo = sp3[2].split()[1]
    
    while True:
        if search in sp3[line]:
            if search in sp3[line+1]:
                line +=1
            else:
                line +=1
                break
        else:
            line +=1
    del sp3[0:line]
    sp3 = [i.replace('P  ', 'PG0') for i in sp3]
    sp3 = [i.replace('P ', 'PG') for i in sp3]
    header = ['X', 'Y', 'Z', 'deltaT', 'sigmaX', 'sigmaY', 'sigmaZ', 'sigmadeltaT', 'Epoch']
    sat, pos = [], []
    while True:
        for i in range(int(SVNo)+1):
            if '*' in sp3[i]:
                sp3[i] = sp3[i].split()
                epoch=datetime.datetime(year = int(sp3[i][1]),
                                        month = int(sp3[i][2]),
                                        day = int(sp3[i][3]),
                                        hour = int(sp3[i][4]),
                                        minute =int(sp3[i][5]),
                                        second = 0)
            else:
                if '999999.999999' in sp3[i]:
                    sp3[i] = sp3[i].replace(' 999999.999999', '          None')
                else:
                    if sp3[i][60:].isspace() == True:
                        sp3[i] = sp3[i][:60]
                    if sp3[i][60:63].isspace() == True: 
                        sp3[i] = sp3[i][:60] + " XX" + sp3[i][63:] 
                    if sp3[i][63:66].isspace() == True: 
                        sp3[i] = sp3[i][:63] + " XX" + sp3[i][66:]  
                    if sp3[i][66:69].isspace() == True: 
                        sp3[i] = sp3[i][:66] + " XX" + sp3[i][69:]  
                    if sp3[i][69:73].isspace() == True: 
                        sp3[i] = sp3[i][:69] + "  XX" + sp3[i][73:] 
                # -------------------------------------------
                sp3[i] = sp3[i].split()
                if len(sp3[i][1:]) == 4:
                    sp3[i].extend(['None', 'None', 'None', 'None'])
                sat.append(sp3[i][0][1:])
                sp3[i] = [float(j) if isfloat(j) == True else None for j in sp3[i]]
                sp3[i].append(epoch)
                pos.append(sp3[i][1:])
        del sp3[0:int(SVNo)+1]
        if 'EOF' in sp3[0]:
            break
    position = pd.DataFrame(pos, index = sat, columns = header)
    position.index.name = 'SV'
    position.set_index('Epoch', append=True, inplace=True)
    position = position.reorder_levels(["Epoch","SV"])
    f.close() # close the file
    # ------------------------------------------------------------------
    end = time.time()
    print('{}'.format(sp3file), 'file is read in', '{0:.2f}'.format(end-start), 'seconds')
    return position

#-------------------------------------------------------------------------
#-------------------------------------------------------------------------
#----------------------------- CLOCK FILE --------------------------------
#-------------------------------------------------------------------------
#-------------------------------------------------------------------------
def read_clockFile(clkFile):
    """ Read Clock file """
    isexist(clkFile)
    start = time.time()
    f = open(clkFile)
    clk = f.readlines()
    line = 0
    prnlist = []
    while True:
        if 'OF SOLN SATS' not in clk[line]:
            del clk[line]
        else:   
            noprn = int(clk[line][4:6])
            line +=1
            break
    while True:
        if 'PRN LIST' in clk[line]:
            prnlist.extend(clk[line])
            prnlist = ''.join(prnlist)
            prnlist= prnlist.split()
            prnlist.remove('PRN')
            prnlist.remove('LIST')
            line +=1
        else:
            break
    SV = []
    for a in range(0, len(prnlist[0]),3):
        SV.extend([prnlist[0][a:a+3]])
        
    if len(prnlist) > 1:
        SV.extend(prnlist[1:])
    
    line = 0
    while True:
            if 'END OF HEADER' not in clk[line]:
                line +=1
            else: 
                del clk[0:line+1]
                break

    timelist = []
    for i in range(len(clk)):
        if clk[i][0:2]=='AS':
            timelist.append(clk[i].split())
    Sat = []
    Epochlist = []
    SVtime = []
    for i in range(len(timelist)):
        Sat.append(timelist[i][1])
        Epochlist.append((datetime.datetime(year = int(timelist[i][2]), month = int(timelist[i][3]),
                                        day = int(timelist[i][4]), hour = int(timelist[i][5]), 
                                        minute = int(timelist[i][6]), second =int(float(timelist[i][7])))))
        SVtime.append(float(timelist[i][9]))
    SVTimelist = pd.DataFrame(list(zip(Epochlist, SVtime)), index = Sat,
                columns=['Epoch','DeltaTSV'])
    f.close() # close the file
    end = time.time()
    print('{}'.format(clkFile), 'file is read in', '{0:.2f}'.format(end-start), 'seconds')
    return SVTimelist
#-------------------------------------------------------------------------
#-------------------------------------------------------------------------
#----------------------------- Ionosphere FILE ---------------------------
#-------------------------------------------------------------------------
#-------------------------------------------------------------------------
def read_ionFile(IonFile):
    """ Function that reads Ionosphere file """
    # check if observationFile exists or not
    isexist(IonFile)
    # ----------------------------------------------
    start = time.time()  
    f = open(IonFile, errors = 'ignore')
    obsLines = f.readlines() 

    line = 0
    while True:
        if 'END OF HEADER' in obsLines[line]:
            line +=1
            break
        else:
            line +=1
    
    del obsLines[0:line]
    # --------------------------------------------------------------------------------------
    tecuList = np.zeros([13, 71,72])
    for etime in range(13):
        line = 0
        del obsLines[0] # start of tec map
        epochLine = obsLines[0].split()
        epoch = datetime.datetime(year = int(epochLine[0]), month = int(epochLine[1]), day = int(epochLine[2]), 
                                            hour = int(epochLine[3]), minute = int(epochLine[4]), second = int(epochLine[5]))
        del obsLines[0] # delete epoch line
        for phi in range(71):
            temp = obsLines[0].split()
            LAT, LON1, LON2, DLON, H = temp[0], temp[1], temp[2], temp[3], temp[4]
            tecu = obsLines[1] + obsLines[2] + obsLines[3] + obsLines[4] + obsLines[5]
            tecu = tecu.split()
            tecu = [int(tec) for tec in tecu]
            for lamda in range(len(tecu)-1):
                tecuList[etime,phi,lamda] = tecu[lamda]
            del obsLines[0:6]
        del obsLines[0]
        #---------------------------------------------------------------------------------------
    f.close() # close the file
    finish = time.time()
    print("Ionosphere file ", IonFile," is read in", "{0:.2f}".format(finish-start), "seconds.")
    return tecuList
